"""Test suite for the RHSSteelProfile class."""

from unittest.mock import MagicMock

import pytest
from matplotlib import pyplot as plt
from matplotlib.figure import Figure

from blueprints.codes.eurocode.en_1993_1_1_2005.chapter_3_materials.table_3_1 import SteelStrengthClass
from blueprints.materials.steel import SteelMaterial
from blueprints.structural_sections.steel.steel_cross_sections.rhs_profile import RHSSteelProfile
from blueprints.structural_sections.steel.steel_cross_sections.standard_profiles.rhs import RHS


class TestRHSSteelProfile:
    """Test suite for RHSSteelProfile."""

    def test_name(self, rhs_profile: RHSSteelProfile) -> None:
        """Test the name of the RHS profile."""
        expected_name: str = "RHS400x200x16"
        assert rhs_profile.name == expected_name

    def test_code(self, rhs_profile: RHSSteelProfile) -> None:
        """Test the code of the RHS profile."""
        expected_alias: str = "RHS400x200x16"
        assert rhs_profile.name == expected_alias

    def test_steel_volume_per_meter(self, rhs_profile: RHSSteelProfile) -> None:
        """Test the steel volume per meter."""
        expected_volume: float = 0.017900  # m³/m
        assert pytest.approx(rhs_profile.volume_per_meter, rel=1e-2) == expected_volume

    def test_steel_weight_per_meter(self, rhs_profile: RHSSteelProfile) -> None:
        """Test the steel weight per meter."""
        expected_weight: float = 0.017900 * 7850  # kg/m
        assert pytest.approx(rhs_profile.weight_per_meter, rel=1e-2) == expected_weight

    def test_area(self, rhs_profile: RHSSteelProfile) -> None:
        """Test the steel cross-sectional area."""
        expected_area: float = 17900  # mm²
        assert pytest.approx(rhs_profile.area, rel=1e-2) == expected_area

    @pytest.mark.slow
    def test_plot(self, rhs_profile: RHSSteelProfile) -> None:
        """Test the plot method (ensure it runs without errors)."""
        fig: Figure = rhs_profile.plot()
        assert isinstance(fig, plt.Figure)

    def test_plot_mocked(self, rhs_profile: RHSSteelProfile, mock_section_properties: MagicMock) -> None:  # noqa: ARG002
        """Test the plotting of the RHS profile shapes with mocked section properties."""
        fig: Figure = rhs_profile.plot()
        assert isinstance(fig, plt.Figure)

    def test_geometry(self, rhs_profile: RHSSteelProfile) -> None:
        """Test the geometry of the RHS profile."""
        expected_geometry = rhs_profile.geometry
        assert expected_geometry is not None

    def test_get_profile_with_corrosion(self) -> None:
        """Test the RHS profile with corrosion applied."""
        # Ensure the profile raises an error if fully corroded
        with pytest.raises(ValueError, match="The profile has fully corroded."):
            RHSSteelProfile.from_standard_profile(
                profile=RHS.RHS400x200_16,
                steel_material=SteelMaterial(SteelStrengthClass.S355),
                corrosion_outside=16,  # mm
                corrosion_inside=0,  # mm
            )

    def test_corrosion_in_name(self) -> None:
        """Test that the name includes corrosion information."""
        rhs_profile_with_corrosion = RHSSteelProfile.from_standard_profile(
            profile=RHS.RHS400x200_16,
            steel_material=SteelMaterial(SteelStrengthClass.S355),
            corrosion_outside=2,  # mm
            corrosion_inside=1,  # mm
        )
        expected_name_with_corrosion = "RHS400x200x16 (corrosion in: 1 mm, out: 2 mm)"
        assert rhs_profile_with_corrosion.name == expected_name_with_corrosion



"""Testing formula 5.31 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_5_structural_analysis.formula_5_31 import Form5Dot31DesignMoment


class TestForm5Dot31DesignMoment:
    """Validation for formula 5.31 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        m_0ed = 50.0
        m_2 = 10.0

        # Object to test
        formula = Form5Dot31DesignMoment(m_0ed=m_0ed, m_2=m_2)

        # Expected result, manually calculated
        manually_calculated_result = 60.0  # kNm

        assert formula == pytest.approx(manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"M_{Ed} = M_{0Ed} + M_{2} = 50.000 + 10.000 = 60.000 \ kNm",
            ),
            ("short", r"M_{Ed} = 60.000 \ kNm"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        m_0ed = 50.0
        m_2 = 10.0

        # Object to test
        latex = Form5Dot31DesignMoment(m_0ed=m_0ed, m_2=m_2).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Circular reinforced cross-section."""

from matplotlib import pyplot as plt
from numpy import cos, pi, sin
from shapely import LineString, Polygon

from blueprints.geometry.operations import rotate_linearring
from blueprints.materials.concrete import ConcreteMaterial
from blueprints.materials.reinforcement_steel import ReinforcementSteelMaterial
from blueprints.structural_sections.concrete.covers import DEFAULT_COVER
from blueprints.structural_sections.concrete.reinforced_concrete_sections.base import ReinforcedCrossSection
from blueprints.structural_sections.concrete.reinforced_concrete_sections.plotters.circular import CircularCrossSectionPlotter
from blueprints.structural_sections.concrete.reinforced_concrete_sections.reinforcement_configurations import ReinforcementByQuantity
from blueprints.structural_sections.concrete.stirrups import StirrupConfiguration
from blueprints.structural_sections.cross_section_circle import CircularCrossSection
from blueprints.type_alias import DEG, DIMENSIONLESS, MM, RATIO


class CircularReinforcedCrossSection(ReinforcedCrossSection):
    """Representation of a reinforced circular concrete cross-section like a column.

    Parameters
    ----------
    diameter : MM
        The diameter of the circular cross-section [mm].
    concrete_material : ConcreteMaterial
        Material properties of the concrete.
    cover : MM, optional
        The reinforcement cover for the cross-section [mm]. The default is 50 mm.
    """

    def __init__(self, diameter: MM, concrete_material: ConcreteMaterial, cover: MM = DEFAULT_COVER) -> None:
        """Initialize the circular reinforced concrete section."""
        super().__init__(
            cross_section=CircularCrossSection(
                diameter=diameter,
                x=0,  # x=0 and y=0 to place the cross-section at the origin
                y=0,
            ),
            concrete_material=concrete_material,
        )
        self.diameter = diameter
        self.cover = cover
        self.plotter = CircularCrossSectionPlotter(cross_section=self)

    def add_stirrup_along_perimeter(
        self,
        diameter: MM,
        distance: MM,
        material: ReinforcementSteelMaterial,
        shear_check: bool = True,
        torsion_check: bool = True,
        mandrel_diameter_factor: DIMENSIONLESS | None = None,
        anchorage_length: MM = 0.0,
        relative_start_position: RATIO = 0.0,
        relative_end_position: RATIO = 1.0,
    ) -> StirrupConfiguration:
        """Adds a stirrup configuration along the perimeter of the cross-section taking the covers into account. The created configuration goes around
        the longitudinal rebars (if any).

        Use .add_stirrup_configuration() to add a stirrup configuration of any shape, size, and position (as long as it is inside the cross-section).

        Parameters
        ----------
        diameter: MM
            Diameter of the stirrups [mm].
        distance: MM
            Longitudinal distance between stirrups [mm].
        material : ReinforcementSteelMaterial
            Representation of the properties of reinforcement steel suitable for use with NEN-EN 1992-1-1
        shear_check: bool
            Take stirrup into account in shear check
        torsion_check: bool
            Take stirrup into account in torsion check
        mandrel_diameter_factor: DIMENSIONLESS
            Inner diameter of mandrel as multiple of stirrup diameter [-]
            (default: 4⌀ for ⌀<=16mm and 5⌀ for ⌀>16mm) Tabel 8.1Na NEN-EN 1992-1-1 Dutch National Annex.
        anchorage_length: MM
            Anchorage length [mm]
        relative_start_position: RATIO
            Relative position of the start of the stirrup configuration inside the cross-section (longitudinal direction). Value between 0 and 1.
            Default is 0 (start).
        relative_end_position: RATIO
            Relative position of the end of the stirrup configuration inside the cross-section (longitudinal direction). Value between 0 and 1.
            Default is 1 (end).

        Returns
        -------
        StirrupConfiguration
            Newly created stirrup configuration inside the cross-section.
        """
        # create the stirrup configuration based on the covers present
        radius = self.diameter / 2 - self.cover - (diameter / 2)
        stirrup_geometry = Polygon([(radius * cos(angle), radius * sin(angle)) for angle in [2 * pi * i / 100 for i in range(100)]])

        # add the stirrup configuration
        return self.add_stirrup_configuration(
            StirrupConfiguration(
                geometry=stirrup_geometry,
                diameter=diameter,
                distance=distance,
                material=material,
                shear_check=shear_check,
                torsion_check=torsion_check,
                mandrel_diameter_factor=mandrel_diameter_factor,
                anchorage_length=anchorage_length,
                based_on_cover=True,
                relative_start_position=relative_start_position,
                relative_end_position=relative_end_position,
            )
        )

    def _get_reference_line(
        self,
        diameter: MM,
        cover: MM | None = None,
        start_angle: DEG = 90.0,
    ) -> LineString:
        """Get the reference circle for the cross-section.

        Parameters
        ----------
        diameter: MM
            Diameter of the rebars [mm].
        cover: MM, optional
            Cover of the rebars [mm]. If not provided, the default cover is used.
        start_angle: DEG
            Starting position of the reference line in degrees. 90 degrees is the top of the circle.
            0 degrees is the right side of the circle. 180 degrees is the right side of the circle.
            Default is 90 degrees.

        Returns
        -------
        LineString
            Reference circle for the cross-section.
        """
        # calculate the effective radius for the rebars
        cover = cover if cover is not None else self.cover

        # check if there is a stirrup configuration present to adjust the radius
        max_stirrups_diameter = 0.0
        if self._stirrups:
            max_stirrups_diameter = max([stirrup.diameter for stirrup in self._stirrups])

        radius = self.diameter / 2 - cover - max_stirrups_diameter - (diameter / 2)

        # create the circle using shapely's Point and buffer
        circle = self.cross_section.centroid.buffer(radius, quad_segs=360, join_style="round")

        return rotate_linearring(linearring=circle.exterior, angle_degrees=start_angle)

    def add_longitudinal_reinforcement_by_quantity(
        self,
        n: int,
        diameter: MM,
        material: ReinforcementSteelMaterial,
        cover: MM | None = None,
        start_angle: DEG = 90.0,
    ) -> None:
        """Add longitudinal reinforcement to the cross-section based on the quantity configuration of rebars.

        Parameters
        ----------
        n: int
            Amount of longitudinal bars.
        diameter: MM
            Diameter of the rebars [mm].
        material : ReinforcementSteelMaterial
            Representation of the properties of reinforcement steel suitable for use with NEN-EN 1992-1-1.
        cover: MM, optional
            Cover of the rebars [mm]. If not provided, the default cover is used.
        start_angle: DEG
            Starting position of the first rebar. Default is 90 degrees which is the top of the circle.
            0 degrees is the right side of the circle. 180 degrees is the right side of the circle.
        """
        return self.add_reinforcement_configuration(
            configuration=ReinforcementByQuantity(
                diameter=diameter,
                material=material,
                n=n,
            ),
            line=self._get_reference_line,
            cover=cover,
            diameter=diameter,
            start_angle=start_angle,
        )

    def plot(self, *args, **kwargs) -> plt.Figure:
        """Plot the cross-section. Making use of the standard plotter.

        If you want to use a custom plotter, use the .plotter attribute to plot the cross-section.

        Parameters
        ----------
        *args
            Additional arguments passed to the plotter.
        **kwargs
            Additional keyword arguments passed to the plotter.
        """
        return self.plotter.plot(*args, **kwargs)



"""Testing formula 3.4 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_3_materials.formula_3_4 import Form3Dot4DevelopmentTensileStrength


class TestForm3Dot4DevelopmentTensileStrength:
    """Validation for formula 3.4 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        beta_cc_t = 0.32  # -
        alpha = 2 / 3  # -
        f_ctm = 3.45  # MPa
        form_3_4 = Form3Dot4DevelopmentTensileStrength(beta_cc_t=beta_cc_t, alpha=alpha, f_ctm=f_ctm)

        # Expected result, manually calculated
        manually_calculated_result = 1.614058

        assert form_3_4 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_raise_error_when_negative_alpha_is_given(self) -> None:
        """Test that an error is raised when alpha is negative."""
        # Example values
        beta_cc_t = 0.32  # -> Positive
        alpha = -0.4  # -> Negative
        f_ctm = 3.45  # MPa -> Positive

        with pytest.raises(ValueError):
            Form3Dot4DevelopmentTensileStrength(beta_cc_t=beta_cc_t, alpha=alpha, f_ctm=f_ctm)

    def test_raise_error_when_negative_beta_cc_t_is_given(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        beta_cc_t = -0.32  # - -> Negative
        alpha = 2 / 3  # - -> Equal to 1 or 2/3
        f_ctm = 3.45  # MPa -> Positive

        with pytest.raises(ValueError):
            Form3Dot4DevelopmentTensileStrength(beta_cc_t=beta_cc_t, alpha=alpha, f_ctm=f_ctm)

    def test_raise_error_when_negative_f_ctm_is_given(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        beta_cc_t = 0.32  # - -> Positive
        alpha = 2 / 3  # - -> unequal to 1 or 2/3
        f_ctm = -3.45  # MPa -> Negative

        with pytest.raises(ValueError):
            Form3Dot4DevelopmentTensileStrength(beta_cc_t=beta_cc_t, alpha=alpha, f_ctm=f_ctm)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"f_{ctm}(t) = (\beta_{cc}(t))^{\alpha} \cdot f_{ctm} = (0.320)^{0.667} \cdot 3.450 = 1.614",
            ),
            ("short", r"f_{ctm}(t) = 1.614"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        beta_cc_t = 0.32  # -
        alpha = 2 / 3  # -
        f_ctm = 3.45  # MPa

        # Object to test
        form_3_4_latex = Form3Dot4DevelopmentTensileStrength(beta_cc_t=beta_cc_t, alpha=alpha, f_ctm=f_ctm).latex()

        actual = {"complete": form_3_4_latex.complete, "short": form_3_4_latex.short}

        assert actual[representation] == expected, f"{representation} representation failed."



"""Annular sector cross-section shape."""

import math
from dataclasses import dataclass

from sectionproperties.pre import Geometry
from shapely.affinity import rotate
from shapely.geometry import Point, Polygon

from blueprints.structural_sections._cross_section import CrossSection
from blueprints.type_alias import DEG, MM


@dataclass(frozen=True)
class AnnularSectorCrossSection(CrossSection):
    """
    Class to represent an annular sector cross-section using shapely for geometric calculations.

    Parameters
    ----------
    inner_radius : MM
        The radius of the inner circle of the annular sector [mm].
    thickness : MM
        The thickness of the annular sector cross-section [mm].
    start_angle : DEG
        The start angle of the annular sector in degrees (top = 0 degrees, clockwise is positive).
    end_angle : DEG
        The end angle of the annular sector in degrees (must be larger than start angle but not more than 360 degrees more).
    x : MM
        The x-coordinate of the annular sector's radius center.
    y : MM
        The y-coordinate of the annular sector's radius center.
    name : str
        The name of the rectangular cross-section, default is "Annular Sector".
    """

    inner_radius: MM
    thickness: MM
    start_angle: DEG
    end_angle: DEG
    x: MM
    y: MM
    name: str = "Annular Sector"

    def __post_init__(self) -> None:
        """Post-initialization to validate the inputs."""
        if self.inner_radius < 0:
            raise ValueError(f"Radius must be zero or positive, but got {self.inner_radius}")
        if self.thickness <= 0:
            raise ValueError(f"Thickness must be a positive value, but got {self.thickness}")
        if self.start_angle > 360 or self.start_angle < -360:
            raise ValueError(f"Start angle must be between -360 and 360 degrees, but got {self.start_angle}")
        if self.end_angle <= self.start_angle:
            raise ValueError(f"End angle must be greater than start angle, but got end angle {self.end_angle} and start angle {self.start_angle}")
        if self.end_angle - self.start_angle >= 360:
            raise ValueError(
                f"The total angle made between start and end angle must be less than 360 degrees, but got "
                f"{self.end_angle - self.start_angle} degrees (end {self.end_angle} - start {self.start_angle})\n\n"
                f"In case you want to create a full circle (donut shape), "
                "use a tube cross section instead (TubeCrossSection)."
            )

    @property
    def radius_centerline(self) -> MM:
        """Calculate the inner radius of the annular sector [mm]."""
        return self.inner_radius + self.thickness / 2.0

    @property
    def outer_radius(self) -> MM:
        """Calculate the outer radius of the annular sector [mm]."""
        return self.radius_centerline + self.thickness / 2.0

    @property
    def height(self) -> MM:
        """
        Calculate the height of the annular sector cross-section [mm].

        Returns
        -------
        MM
            The height of the annular sector.
        """
        min_y = min(y for _, y in self.polygon.exterior.coords)
        max_y = max(y for _, y in self.polygon.exterior.coords)
        return max_y - min_y

    @property
    def width(self) -> MM:
        """
        Calculate the width of the annular sector cross-section [mm].

        Returns
        -------
        MM
            The width of the annular sector.
        """
        min_x = min(x for x, _ in self.polygon.exterior.coords)
        max_x = max(x for x, _ in self.polygon.exterior.coords)
        return max_x - min_x

    @property
    def polygon(self) -> Polygon:
        """
        Shapely Polygon representing the annular sector cross-section.

        Returns
        -------
        Polygon
            The shapely Polygon representing the annular sector.
        """
        center = Point(self.x, self.y)
        inner_circle = center.buffer(self.inner_radius, quad_segs=64)
        outer_circle = center.buffer(self.outer_radius, quad_segs=64)

        inner_ring = rotate(inner_circle, -self.start_angle, origin=center)
        outer_ring = rotate(outer_circle, -self.start_angle, origin=center)

        # Create the annular sector by intersecting with a sector
        sector_points = [center]
        angle_step = (self.end_angle - self.start_angle) / 8
        for i in range(9):
            angle = math.radians(90 - self.start_angle - i * angle_step)
            sector_points.append(Point(center.x + 2 * self.outer_radius * math.cos(angle), center.y + 2 * self.outer_radius * math.sin(angle)))
        sector_points.append(center)
        sector = Polygon(sector_points).buffer(0)

        result = outer_ring.difference(inner_ring).intersection(sector)
        return Polygon(result)  # type: ignore[arg-type]

    def geometry(
        self,
        mesh_size: MM | None = None,
    ) -> Geometry:
        """Return the geometry of the annular sector cross-section.

        Parameters
        ----------
        mesh_size : MM
            Maximum mesh element area to be used within
            the Geometry-object finite-element mesh. If not provided, a default value will be used.

        Returns
        -------
        Geometry
            The Geometry object representing the annular sector.
        """
        if mesh_size is None:
            minimum_mesh_size = 1.0
            mesh_length = max(self.thickness / 5, minimum_mesh_size)
            mesh_size = mesh_length**2

        annular_sector = Geometry(geom=self.polygon)
        annular_sector.create_mesh(mesh_sizes=mesh_size)
        return annular_sector



"""Testing formula 12.3 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_12_plain_and_lightly_reinforced_concrete_structures.formula_12_3 import (
    Form12Dot3PlainConcreteShearStress,
)
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestForm12Dot3PlainConcreteShearStress:
    """Validation for formula 12.3 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        n_ed = 100000.0  # N
        a_cc = 50000.0  # mm^2

        # Object to test
        form_12_3 = Form12Dot3PlainConcreteShearStress(n_ed=n_ed, a_cc=a_cc)

        # Expected result, manually calculated
        manually_calculated_result = 2.0  # MPa

        assert round(form_12_3, 3) == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("n_ed", "a_cc"),
        [
            (100000.0, 0.0),
        ],
    )
    def test_raise_error_when_values_are_less_or_equal_to_zero(
        self,
        n_ed: float,
        a_cc: float,
    ) -> None:
        """Test values less or equal to zero for n_ed and a_cc."""
        with pytest.raises(LessOrEqualToZeroError):
            Form12Dot3PlainConcreteShearStress(n_ed=n_ed, a_cc=a_cc)

    @pytest.mark.parametrize(
        ("n_ed", "a_cc"),
        [
            (-100000.0, 500),
        ],
    )
    def test_raise_error_when_values_are_less_to_zero(
        self,
        n_ed: float,
        a_cc: float,
    ) -> None:
        """Test values less or equal to zero for n_ed and a_cc."""
        with pytest.raises(NegativeValueError):
            Form12Dot3PlainConcreteShearStress(n_ed=n_ed, a_cc=a_cc)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"\sigma_{cp} = \frac{N_{Ed}}{A_{cc}} = \frac{100000.000}{50000.000} = 2.000",
            ),
            ("short", r"\sigma_{cp} = 2.000"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        n_ed = 100000.0  # N
        a_cc = 50000.0  # mm^2

        # Object to test
        form_12_3_latex = Form12Dot3PlainConcreteShearStress(
            n_ed=n_ed,
            a_cc=a_cc,
        ).latex()

        actual = {
            "complete": form_12_3_latex.complete,
            "short": form_12_3_latex.short,
            "string": str(form_12_3_latex),
        }

        assert actual[representation] == expected, f"{representation} representation failed."



"""Formula 9.11 from EN 1992-1-1:2004: Chapter 9 - Detailing and specific rules."""

import numpy as np

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula, latex_replace_symbols
from blueprints.type_alias import DEG, MM, MM2, MPA
from blueprints.validations import raise_if_greater_than_90, raise_if_less_or_equal_to_zero, raise_if_negative


class Form9Dot11MinimumShearReinforcement(Formula):
    """Class representing the formula 9.11 for the calculation of the minimum shear reinforcement."""

    label = "9.11"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        alpha: DEG,
        s_r: MM,
        s_t: MM,
        f_ck: MPA,
        f_yk: MPA,
    ) -> None:
        r"""[$$A_{sw,min}$$] Minimum shear reinforcement [$$mm^2$$].

        The spirit of the equation is to calculate the shear reinforcement area. The formula is converted such that
        it actully does that, as opposed to the presentation in the Eurocode which only checks if the area is sufficient
        with a boolean result.

        EN 1992-1-1:2004 art.9.4.3(2) - Formula (9.11)

        Parameters
        ----------
        alpha : DEG
            [$$\alpha$$] Angle between the shear reinforcement and the main steel [$$^\circ$$].
        s_r : MM
            [$$s_r$$] Spacing of shear links in the radial direction [$$mm$$].
        s_t : MM
            [$$s_t$$] Spacing of shear links in the tangential direction [$$mm$$].
        f_ck : MPA
            [$$f_{ck}$$] Characteristic compressive strength of concrete [$$MPa$$].
        f_yk : MPA
            [$$f_{yk}$$] Characteristic yield strength of reinforcement [$$MPa$$].
        """
        super().__init__()
        self.alpha = alpha
        self.s_r = s_r
        self.s_t = s_t
        self.f_ck = f_ck
        self.f_yk = f_yk

    @staticmethod
    def _evaluate(
        alpha: DEG,
        s_r: MM,
        s_t: MM,
        f_ck: MPA,
        f_yk: MPA,
    ) -> MM2:
        """For more detailed documentation see the class docstring."""
        raise_if_less_or_equal_to_zero(s_r=s_r, s_t=s_t, f_yk=f_yk)
        raise_if_negative(alpha=alpha, f_ck=f_ck)
        raise_if_greater_than_90(alpha=alpha)

        alpha_rad = np.deg2rad(alpha)
        lhs = (1.5 * np.sin(alpha_rad) + np.cos(alpha_rad)) / (s_r * s_t)
        rhs = 0.08 * np.sqrt(f_ck) / f_yk

        return rhs / lhs

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 9.11."""
        _equation: str = (
            r"\frac{0.08 \cdot \sqrt{f_{ck}}}{f_{yk}} \cdot \frac{s_r \cdot s_t}{1.5 \cdot "
            r"\sin(\alpha) + \cos(\alpha)}"
        )
        _numeric_equation: str = latex_replace_symbols(
            _equation,
            {
                r"\alpha": f"{self.alpha:.{n}f}",
                "s_r": f"{self.s_r:.{n}f}",
                "s_t": f"{self.s_t:.{n}f}",
                "f_{ck}": f"{self.f_ck:.{n}f}",
                "f_{yk}": f"{self.f_yk:.{n}f}",
            },
            False,
        )
        return LatexFormula(
            return_symbol=r"A_{sw,min}",
            result=f"{self:.{n}f}",
            equation=_equation,
            numeric_equation=_numeric_equation,
            comparison_operator_label="=",
            unit="mm^2",
        )



"""Testing formula 3.12 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_3_materials.formula_3_12 import Form3Dot12AutogeneShrinkageInfinity


class TestForm3Dot11AutogeneShrinkage:
    """Validation for formula 3.12 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        f_ck = 15.8  # MPa
        form_3_12 = Form3Dot12AutogeneShrinkageInfinity(f_ck=f_ck)

        # Expected result, manually calculated
        manually_calculated_result = 1.45e-5

        assert form_3_12 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_raise_error_when_f_ck_is_given(self) -> None:
        """Test formula raising error by a negative value."""
        # Example values
        f_ck = -15.8  # MPa

        with pytest.raises(ValueError):
            Form3Dot12AutogeneShrinkageInfinity(f_ck=f_ck)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"\epsilon_{ca}(\infty) = 2.5 \cdot (f_{ck} - 10) \cdot 10^{-6} = 2.5 \cdot (15.800 - 10) \cdot 10^{-6} = 0.000015",
            ),
            ("short", r"\epsilon_{ca}(\infty) = 0.000015"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        f_ck = 15.8  # MPa

        # Object to test
        form_3_12_latex = Form3Dot12AutogeneShrinkageInfinity(f_ck=f_ck).latex()

        actual = {"complete": form_3_12_latex.complete, "short": form_3_12_latex.short}

        assert actual[representation] == expected, f"{representation} representation failed."



"""Testing formula 4.1 of NEN-EN 1992-1-1:2005+A1:2015+NB:2016+A1:2020."""

import pytest

from blueprints.codes.eurocode.nen_en_1992_1_1_a1_2020.chapter_4_durability_and_cover.formula_4_1 import Form4Dot1NominalConcreteCover
from blueprints.validations import NegativeValueError


class TestForm4Dot1NominalConcreteCover:
    """Validation for formula 4.1 from NEN-EN 1992-1-1:2005+A1:2015+NB:2016+A1:2020."""

    def test_evaluation(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        c_min = 60  # mm
        delta_c_dev = 5  # mm
        form_4_1 = Form4Dot1NominalConcreteCover(c_min=c_min, delta_c_dev=delta_c_dev)

        # Expected result, manually calculated
        manually_calculated_result = 65

        assert form_4_1 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_raise_error_when_negative_c_min_is_given(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        c_min = -60  # mm
        delta_c_dev = 5  # mm

        with pytest.raises(NegativeValueError):
            Form4Dot1NominalConcreteCover(c_min=c_min, delta_c_dev=delta_c_dev)

    def test_raise_error_when_negative_delta_c_dev_is_given(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        c_min = 60  # mm
        delta_c_dev = -5  # mm

        with pytest.raises(NegativeValueError):
            Form4Dot1NominalConcreteCover(c_min=c_min, delta_c_dev=delta_c_dev)

    @pytest.mark.parametrize(
        ("representation", "expected_result"),
        [
            ("complete", r"c_{nom} = c_{min}+\Delta c_{dev} = 60+5 = 65.0"),
            ("short", "c_{nom} = 65.0"),
            ("string", r"c_{nom} = c_{min}+\Delta c_{dev} = 60+5 = 65.0"),
        ],
    )
    def test_latex(self, representation: str, expected_result: str) -> None:
        """Test the latex implementation."""
        c_min = 60  # mm
        delta_c_dev = 5  # mm
        form = Form4Dot1NominalConcreteCover(c_min=c_min, delta_c_dev=delta_c_dev).latex()

        actual = {
            "complete": form.complete,
            "short": form.short,
            "string": str(form),
        }

        assert actual[representation] == expected_result, f"{representation} representation failed."



"""Testing formula 5.7 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_5_structural_analysis.formula_5_7 import Form5Dot7EffectiveFlangeWidth
from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_5_structural_analysis.formula_5_7ab import Form5Dot7abFlangeEffectiveFlangeWidth
from blueprints.validations import NegativeValueError


class TestForm5Dot7EffectiveFlangeWidth:
    """Validation for formula 5.7 from EN 1992-1-1:2004."""

    @pytest.mark.parametrize(
        ("b", "b_eff"),
        [
            (0.8, 0.6),
            (0.55, 0.55),
        ],
    )
    def test_evaluation(self, b: float, b_eff: float) -> None:
        """Test the evaluation of the result."""
        # Example values
        b_eff_i = [0.2, 0.25]  # M
        b_w = 0.15  # M

        # Object to test
        form_5_7 = Form5Dot7EffectiveFlangeWidth(*b_eff_i, b_w=b_w, b=b)

        # Expected result, manually calculated
        manually_calculated_result = b_eff  # M

        assert form_5_7 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("b_eff_i", "b_w", "b"),
        [
            ([0.2, -0.25], 5, 0.8),
            ([-0.2, -0.25], 5, 0.8),
            ([0.2, 0.25], -5, 0.8),
            ([0.2, 0.25], 5, -0.8),
            ([-0.2, -0.25], -5, -0.8),
        ],
    )
    def test_raise_error_when_negative_values_are_given(self, b_eff_i: list[float], b_w: float, b: float) -> None:
        """Test negative values."""
        with pytest.raises(NegativeValueError):
            Form5Dot7EffectiveFlangeWidth(*b_eff_i, b_w=b_w, b=b)

    def test_integration_of_formulas_5_7_and_5_7a_and_5_7b(self) -> None:
        """Test the integration of formulas 5.7a and 5.7b with formula 5.7."""
        # Example values
        b_1 = 0.25  # M
        b_2 = 0.5  # M
        l_0 = 1.5  # M
        b = 0.8  # M
        b_w = 0.15  # M

        b_eff_i = (
            Form5Dot7abFlangeEffectiveFlangeWidth(b_1, l_0),
            Form5Dot7abFlangeEffectiveFlangeWidth(b_2, l_0),
        )
        # Object to test
        form_5_7 = Form5Dot7EffectiveFlangeWidth(*b_eff_i, b_w=b_w, b=b)

        # Expected result, manually calculated
        manually_calculated_result = 0.6  # M

        assert form_5_7 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"b_{eff} = \sum b_{eff,i}+b_w\le b = \sum (0.2+0.25)+0.15\le 0.75 = 0.600",
            ),
            ("short", r"b_{eff} = 0.600"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        b_eff_i = [0.2, 0.25]  # M
        b_w = 0.15  # M
        b = 0.75  # M

        # Object to test
        form_5_7_latex = Form5Dot7EffectiveFlangeWidth(*b_eff_i, b_w=b_w, b=b).latex()

        actual = {
            "complete": form_5_7_latex.complete,
            "short": form_5_7_latex.short,
        }

        assert expected == actual[representation], f"{representation} representation."



"""Formula 5.2 from EN 1992-1-1:2004: Chapter 5 - Structural Analysis."""

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DIMENSIONLESS, M
from blueprints.validations import raise_if_less_or_equal_to_zero, raise_if_negative


class Form5Dot2Eccentricity(Formula):
    """Class representing formula 5.2 for the calculation of eccentricity, [$e_i$]."""

    label = "5.2"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        theta_i: DIMENSIONLESS,
        l_0: M,
    ) -> None:
        r"""[$e_i$] Eccentricity, [$e_i$], for isolated members [$m$].

        EN 1992-1-1:2004 art.5.2(7) - Formula (5.2)

        Parameters
        ----------
        theta_i : DIMENSIONLESS
            [$\Theta_i$] Eccentricity, initial inclination imperfections [-].
            Use your own implementation of this value or use the Form5Dot2Imperfections class.
        l_0 : M
            [$l_0$] Effective length of the member, see 5.8.3.2 [$m$].
        """
        super().__init__()
        self.theta_i = theta_i
        self.l_0 = l_0

    @staticmethod
    def _evaluate(
        theta_i: float,
        l_0: M,
    ) -> float:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(theta_i=theta_i)
        raise_if_less_or_equal_to_zero(l_0=l_0)
        return theta_i * l_0 / 2

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 5.2."""
        return LatexFormula(
            return_symbol=r"e_i",
            result=f"{self:.4f}",
            equation=r"\theta_i \cdot l_0 / 2",
            numeric_equation=rf"{self.theta_i:.{n}f} \cdot {self.l_0:.{n}f} / 2",
            comparison_operator_label="=",
        )



r"""Formula 8.4 from EN 1992-1-1:2004: Chapter 8: Detailing of reinforcement and prestressing tendons."""

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_8_detailing_of_reinforcement_and_prestressing_tendons.formula_8_5 import (
    Form8Dot5ProductAlphas235,
)
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula, latex_max_curly_brackets
from blueprints.type_alias import MM, RATIO
from blueprints.validations import raise_if_negative


class Form8Dot4DesignAnchorageLength(Formula):
    r"""Class representing formula 8.4 for the calculation of the design anchorage length [$l_{bd}$] [$mm$]."""

    label = "8.4"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        alpha_1: RATIO,
        alpha_2: RATIO,
        alpha_3: RATIO,
        alpha_4: RATIO,
        alpha_5: RATIO,
        l_b_rqd: MM,
        l_b_min: MM,
        min_product_alpha_2_3_5: RATIO | None = None,
    ) -> None:
        r"""[$l_{bd}$] Design anchorage length [$mm$].

        EN 1992-1-1:2004 art.8.4.4(1) - Formula (8.4)

        Parameters
        ----------
        alpha_1 : RATIO
            [$α_{1}$] Coefficient for the effect of the form of the bars assuming adequate cover (see figure 8.1) [$-$].
            [$= 1.0$] for bars in compression.
            [$= 1.0$] for straight bars in tension.
            [$= 1.0 \text{ if } c_{d} \leq 3 ⋅ Ø$] for bars other than straight in tension (see figure 8.1 (b), (c) and (d)).
            [$= 0.7 \text{ if } c_{d} > 3 ⋅ Ø$] for bars other than straight in tension (see figure 8.1 (b), (c) and (d)).
            Note: see figure 8.3 for values of [$c_{d}$].
        alpha_2 : RATIO
            [$α_{2}$] Coefficient for the effect of minimum concrete cover (see figure 8.3) [$-$].
            [$= 1.0$] for bars in compression.
            [$= 1 - 0.15 ⋅ (c_{d} - Ø) / Ø \leq 1$] with a minimum of 0.7 for straight bars in tension.
            [$= 1 - 0.15 ⋅ (c_{d} - 3 ⋅ Ø) / Ø \leq 1$] with a minimum of 0.7 for bars other than
            straight in tension (see figure 8.1 (b), (c) and (d)).
            Note: see figure 8.3 for values of [$c_{d}$].
        alpha_3 : RATIO
            [$α_{3}$] Coefficient for the effect of confinement by transverse reinforcement [$-$].
            [$= 1.0$] for bars in compression.
            [$= 1 - K ⋅ λ \leq 1$] with a minimum of 0.7 for bars in tension.
            Where: [$λ = (\Sigma A_{st} - \Sigma A_{st,min}) / A_{s}$].
            Where: [$\Sigma A_{st,min}$] = cross-sectional area of the minimum transverse
            reinforcement [$= 0,25 ⋅ A_{s}$] for beams and 0 for slabs.
            Note: see figure 8.4 for values of [$K, A_{s}$] and [$A_{st}$].
        alpha_4 : RATIO
            [$α_{4}$] Coefficient for the influence of one or more welded transverse bars [$Ø_{t} > 0,6 Ø$] along the design anchorage
            length [$l_{bd}$] (see 8.6) [$-$].
            [$= 0.7$] for all types, position and size as specified in figure 8.6 (e) in both tension and compression.
        alpha_5 : RATIO
            [$α_{5}$] Coefficient for the effect of the pressure transverse to the plane of splitting
            along the design anchorage length [$l_{bd}$] (see 8.6) [$-$].
            [$= 1 - 0.04 ⋅ p \leq 1$] with a minimum of 0.7 for all types of anchorage in compression.
            Where: p = transverse pressure at ultimate limit state along [$l_{bd}$] [$MPa$].
        l_b_rqd: MM
            [$l_{b,rqd}$] Basic required anchorage length, for anchoring the force [$A_{s}⋅σ_{sd}$] in a straight bar assuming constant
            bond stress (formula 8.3) [$mm$].
            Use your own implementation for this value or use the :class:`Form8Dot3RequiredAnchorageLength` class.
        l_b_min : MM
            [$l_{b,min}$] Minimum anchorage length if no other limitation is applied [$mm$].
            [$= \max(0.3 ⋅ l_{b,rqd}, 10 ⋅ Ø, 100)$] for tension anchorage (formula 8.6).
            [$= \max(0.6 ⋅ l_{b,rqd}, 10 ⋅ Ø, 100)$] for compression anchorage (formula 8.7).
            Use your own implementation of this formula or use the :class:`Form8Dot6MinimumTensionAnchorage` class for tension or
            :class:`Form8Dot7MinimumCompressionAnchorage` for compression.
        min_product_alpha_2_3_5: RATIO | None
            Minimum value of the product of factors [$α_{2}$], [$α_{3}$] and [$α_{5}$].
            When this argument is None, :class: `Form8Dot5ProductAlphas235` is used for this condition.
            When this argument is given, the condition [$\max(α_{2}⋅α_{3}⋅α_{5}) \geq$] min_product_alpha_2_3_5 is used.

        Notes
        -----
        EN 1992-1-1:2004 art.8.4.4(1) - Formula (8.5) prescribes that [$α_{2} ⋅ α_{3} ⋅ α_{5} \geq 0.7$].
        """
        super().__init__()
        self.alpha_1 = alpha_1
        self.alpha_2 = alpha_2
        self.alpha_3 = alpha_3
        self.alpha_4 = alpha_4
        self.alpha_5 = alpha_5
        self.l_b_rqd = l_b_rqd
        self.l_b_min = l_b_min
        self.min_product_alpha_2_3_5 = min_product_alpha_2_3_5

    @staticmethod
    def _evaluate(
        alpha_1: RATIO,
        alpha_2: RATIO,
        alpha_3: RATIO,
        alpha_4: RATIO,
        alpha_5: RATIO,
        l_b_rqd: MM,
        l_b_min: MM,
        min_product_alpha_2_3_5: RATIO | None = None,
    ) -> MM:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(
            alpha_1=alpha_1,
            alpha_2=alpha_2,
            alpha_3=alpha_3,
            alpha_4=alpha_4,
            alpha_5=alpha_5,
            l_b_rqd=l_b_rqd,
            l_b_min=l_b_min,
        )

        if min_product_alpha_2_3_5 is None:
            product_alphas_2_3_5: float = Form8Dot5ProductAlphas235(alpha_2=alpha_2, alpha_3=alpha_3, alpha_5=alpha_5)
        else:
            product_alphas_2_3_5 = max(alpha_2 * alpha_3 * alpha_5, min_product_alpha_2_3_5)

        return max(alpha_1 * alpha_4 * product_alphas_2_3_5 * l_b_rqd, l_b_min)

    def latex(self, n: int = 2) -> LatexFormula:
        """Returns a LatexFormula representation of the formula."""
        return LatexFormula(
            return_symbol=r"l_{bd}",
            result=f"{self:.{n}f}",
            equation=latex_max_curly_brackets(r"\alpha_1 \cdot \alpha_2 \cdot \alpha_3 \cdot \alpha_4 \cdot \alpha_5 \cdot l_{b,rqd}", r"l_{b,min}"),
            numeric_equation=latex_max_curly_brackets(
                rf"{self.alpha_1} \cdot {self.alpha_2} \cdot {self.alpha_3} \cdot {self.alpha_4} \cdot {self.alpha_5} \cdot {self.l_b_rqd:.{n}f}",
                self.l_b_min,
            ),
            comparison_operator_label="=",
        )



"""Formula 5.9 from EN 1992-1-1:2004: Chapter 5 - Structural Analysis."""

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import KN, M
from blueprints.validations import raise_if_negative


class Form5Dot9DesignSupportMomentReduction(Formula):
    """Class representing formula 5.9 for the calculation of the design support moment reduction, [$ΔM_{Ed}$].
    See Figure 5.4 b.
    """

    label = "5.9"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        f_ed_sup: KN,
        t: M,
    ) -> None:
        r"""[$ΔM_{Ed}$] Design support moment reduction [$kN$].

        Note: Regardless of the method of analysis used, where a beam or slab is continuous over a
        support which may be considered to provide no restraint to rotation (e.g. over walls), the design
        support moment, calculated on the basis of a span equal to the centre-to-centre distance
        between supports, may be reduced by an amount [$ΔM_{Ed}$].

        EN 1992-1-1:2004 art.5.3.2.2(4) - Formula (5.9)

        Parameters
        ----------
        f_ed_sup : KN
            [$F_{Ed,sup}$] Design support reaction [$kN$].
        t : M
            [$t$] breadth of the support (see Figure 5.4 b) [$m$].

            Note:  Where support bearings are used [$t$] should be taken as the bearing width.
        """
        super().__init__()
        self.f_ed_sup = f_ed_sup
        self.t = t

    @staticmethod
    def _evaluate(
        f_ed_sup: KN,
        t: M,
    ) -> KN:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(
            f_ed_sup=f_ed_sup,
            t=t,
        )
        return f_ed_sup * t / 8

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 5.9."""
        return LatexFormula(
            return_symbol=r"ΔM_{Ed}",
            result=f"{self:.{n}f}",
            equation=r"\frac{F_{Ed,sup} \cdot t}{8}",
            numeric_equation=rf"\frac{{{self.f_ed_sup} \cdot {self.t}}}{{8}}",
            comparison_operator_label="=",
        )



"""Testing formula 3.9 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_3_materials.formula_3_9 import Form3Dot9DryingShrinkage


class TestForm3Dot9DryingShrinkage:
    """Validation for formula 3.9 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        beta_ds_tt_s = 0.25  # -
        k_h = 0.75  # -
        epsilon_cd_0 = 0.44  # -
        form_3_9 = Form3Dot9DryingShrinkage(beta_ds_tt_s=beta_ds_tt_s, k_h=k_h, epsilon_cd_0=epsilon_cd_0)

        # Expected result, manually calculated
        manually_calculated_result = 0.0825

        assert form_3_9 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_raise_error_when_negative_beta_ds_tt_s_is_given(self) -> None:
        """Test that an error is raised when beta_ds_tt_s is negative."""
        # Example values
        beta_ds_tt_s = -0.25  # -
        k_h = 0.75  # -
        epsilon_cd_0 = 0.44  # -

        with pytest.raises(ValueError):
            Form3Dot9DryingShrinkage(beta_ds_tt_s=beta_ds_tt_s, k_h=k_h, epsilon_cd_0=epsilon_cd_0)

    def test_raise_error_when_negative_k_h_is_given(self) -> None:
        """Test that an error is raised when k_h is negative."""
        # Example values
        beta_ds_tt_s = 0.25  # -
        k_h = -0.75  # -
        epsilon_cd_0 = 0.44  # -

        with pytest.raises(ValueError):
            Form3Dot9DryingShrinkage(beta_ds_tt_s=beta_ds_tt_s, k_h=k_h, epsilon_cd_0=epsilon_cd_0)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"\epsilon_{cd}(t) = \beta_{ds}(t,t_s) \cdot k_h \cdot \epsilon_{cd,0} = 0.250 \cdot 0.750 \cdot 0.440 = 0.083",
            ),
            ("short", r"\epsilon_{cd}(t) = 0.083"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        beta_ds_tt_s = 0.25  # -
        k_h = 0.75  # -
        epsilon_cd_0 = 0.44  # -

        # Object to test
        form_3_9_latex = Form3Dot9DryingShrinkage(beta_ds_tt_s=beta_ds_tt_s, k_h=k_h, epsilon_cd_0=epsilon_cd_0).latex()

        actual = {"complete": form_3_9_latex.complete, "short": form_3_9_latex.short}

        assert actual[representation] == expected, f"{representation} representation failed."



"""Test class for the reinforcement steel material object."""

import pytest

from blueprints.materials.reinforcement_steel import ReinforcementSteelMaterial, ReinforcementSteelQuality


class TestReinforcementSteelMaterial:
    """Test class for the reinforcement steel material object."""

    def test_default_name(self, fixture_reinforcement_steel_material_b500b: ReinforcementSteelMaterial) -> None:
        """Tests the name."""
        assert fixture_reinforcement_steel_material_b500b.name == "B500B"

    def test_name(self) -> None:
        """Tests the name."""
        steel_with_given_name = ReinforcementSteelMaterial(
            steel_quality=ReinforcementSteelQuality.B500B,
            custom_name="B500B with a different name",
        )
        assert steel_with_given_name.name == "B500B with a different name"

    def test_e_s(self, fixture_reinforcement_steel_material_b500b: ReinforcementSteelMaterial) -> None:
        """Tests the E_s property."""
        assert fixture_reinforcement_steel_material_b500b.e_s == 200000

    def test_custom_given_e_s(self) -> None:
        """Tests the custom given E_s."""
        steel_with_given_e_s = ReinforcementSteelMaterial(
            steel_quality=ReinforcementSteelQuality.B500B,
            custom_e_s=300000,
        )
        assert steel_with_given_e_s.e_s == 300000

    def test_f_yk(self, fixture_reinforcement_steel_material_b500b: ReinforcementSteelMaterial) -> None:
        """Tests the f_yk property."""
        assert fixture_reinforcement_steel_material_b500b.f_yk == 500

    def test_f_yd(self, fixture_reinforcement_steel_material_b500b: ReinforcementSteelMaterial) -> None:
        """Tests the f_yd property."""
        assert fixture_reinforcement_steel_material_b500b.f_yd == pytest.approx(expected=434.78260869, rel=1e-4)

    def test_steel_class(self, fixture_reinforcement_steel_material_b500b: ReinforcementSteelMaterial) -> None:
        """Tests the steel_class property."""
        assert fixture_reinforcement_steel_material_b500b.steel_class == "B"

    def test_f_tk(self, fixture_reinforcement_steel_material_b500b: ReinforcementSteelMaterial) -> None:
        """Tests the f_tk property."""
        assert fixture_reinforcement_steel_material_b500b.f_tk == 540

    @pytest.mark.parametrize(
        ("steel_quality", "expected"),
        [
            (ReinforcementSteelQuality.B500A, 1.05),
            (ReinforcementSteelQuality.B500B, 1.08),
            (ReinforcementSteelQuality.B500C, 1.15),
        ],
    )
    def test_ductility_factor_k(self, steel_quality: ReinforcementSteelQuality, expected: str) -> None:
        """Tests the ductility_factor_k property."""
        steel_material = ReinforcementSteelMaterial(steel_quality=steel_quality)
        assert steel_material.ductility_factor_k == expected

    def test_ductility_factor_k_raises_value_error(self) -> None:
        """Tests the ductility_factor_k property raises a ValueError when the ReinforcementSteelQuality is invalid."""
        invalid_steel_quality = type("InvalidReinforcementSteelQuality", (), {"value": "invalid_value"})()
        invalid_steel_material = ReinforcementSteelMaterial(steel_quality=invalid_steel_quality)
        with pytest.raises(ValueError):
            invalid_steel_material.ductility_factor_k

    @pytest.mark.parametrize(
        ("steel_quality", "expected"),
        [
            (ReinforcementSteelQuality.B500A, 250),
            (ReinforcementSteelQuality.B500B, 500),
            (ReinforcementSteelQuality.B500C, 750),
        ],
    )
    def test_eps_uk(self, steel_quality: ReinforcementSteelQuality, expected: str) -> None:
        """Tests the eps_uk property."""
        steel_material = ReinforcementSteelMaterial(steel_quality=steel_quality)
        assert steel_material.eps_uk == expected

    def test_eps_uk_raises_value_error(self) -> None:
        """Tests the eps_uk property raises a ValueError when the ReinforcementSteelQuality is invalid."""
        invalid_steel_quality = type("InvalidReinforcementSteelQuality", (), {"value": "invalid_value"})()
        invalid_steel_material = ReinforcementSteelMaterial(steel_quality=invalid_steel_quality)
        with pytest.raises(ValueError):
            invalid_steel_material.eps_uk



"""Test the SHSCF enum."""

from blueprints.structural_sections.steel.steel_cross_sections.standard_profiles.shscf import SHSCF


class TestSHSCF:
    """Tests for the SHSCF enum."""

    def test_enum_values(self) -> None:
        """Test that enum values are correctly defined."""
        assert SHSCF.SHSCF_20_2.value == ("SHSCF20x2", 20, 2, 4, 2)
        assert SHSCF.SHSCF_600_20.value == ("SHSCF600x20", 600, 20, 60, 40)

    def test_enum_membership(self) -> None:
        """Test that specific values are members of the enum."""
        assert "SHSCF20x2" in [e.value[0] for e in SHSCF]
        assert "SHSCF600x20" in [e.value[0] for e in SHSCF]

    def test_enum_uniqueness(self) -> None:
        """Test that all enum values are unique."""
        values = [e.value for e in SHSCF]
        assert len(values) == len(set(values))

    def test_enum_attributes(self) -> None:
        """Test that enum attributes are correctly assigned."""
        profile = SHSCF.SHSCF_20_2
        assert profile.alias == "SHSCF20x2"
        assert profile.total_width == 20
        assert profile.total_height == 20
        assert profile.thickness == 2
        assert profile.outer_radius == 4
        assert profile.inner_radius == 2



"""Formula 12.4 from EN 1992-1-1:2004: Chapter 12 - Plain and Lightly Reinforced Concrete Structures."""

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DIMENSIONLESS, MM2, MPA, N
from blueprints.validations import raise_if_less_or_equal_to_zero, raise_if_negative


class Form12Dot4PlainConcreteShearStress(Formula):
    r"""Class representing formula 12.4 for the calculation of the design shear stress of plain concrete, :math:`\tau_{cp}`.

    EN 1992-1-1:2004 art.12.6.3(3) - Formula (12.4)
    """

    label = "12.4"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        k: DIMENSIONLESS,
        v_ed: N,
        a_cc: MM2,
    ) -> None:
        r"""[:math:`\tau_{cp}`] Design shear stress of plain concrete [:math:`MPa`].

        EN 1992-1-1:2004 art.12.6.3(2) - Formula (12.4)

        Parameters
        ----------
        k : DIMENSIONLESS
            [:math:`k`] Nationally determined parameter, recommended value is 1.5 [-].
        v_ed : N
            [:math:`V_{Ed}`] Design shear force [:math:`N`].
        a_cc : MM2
            [:math:`A_{cc}`] Compressed area [:math:`mm^2`].
        """
        super().__init__()
        self.k = k
        self.v_ed = v_ed
        self.a_cc = a_cc

    @staticmethod
    def _evaluate(
        k: DIMENSIONLESS,
        v_ed: N,
        a_cc: MM2,
    ) -> MPA:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(
            v_ed=v_ed,
            k=k,
        )
        raise_if_less_or_equal_to_zero(
            a_cc=a_cc,
        )
        return k * v_ed / a_cc

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 12.4."""
        return LatexFormula(
            return_symbol=r"\tau_{cp}",
            result=f"{self:.{n}f}",
            equation=r"k \cdot \frac{V_{Ed}}{A_{cc}}",
            numeric_equation=rf"{self.k:.{n}f} \cdot \frac{{{self.v_ed:.{n}f}}}{{{self.a_cc:.{n}f}}}",
            comparison_operator_label="=",
        )


"""Helper class for sigma_cp vs sigma_c,lim evaluation."""


class Form12Dot4PlainConcreteShearStressComparison:
    r"""Class representing the comparison of σcp and σc,lim for plain concrete shear stress."""

    label = "12.4 Comparison"
    source_document = EN_1992_1_1_2004

    def __init__(self, sigma_cp: MPA, sigma_c_lim: MPA) -> None:
        r"""[:math:`σ_{cp} ≤ σ_{c,lim}`] Comparison of design shear stress and limit compressive stress.

        EN 1992-1-1:2004 art.12.6.3

        Parameters
        ----------
        sigma_cp : MPA
            [:math:`σ_{cp}`] Design shear stress [:math:`MPa`].
        sigma_c_lim : MPA
            [:math:`σ_{c,lim}`] Limit compressive stress [:math:`MPa`].
        """
        super().__init__()
        self.sigma_cp = sigma_cp
        self.sigma_c_lim = sigma_c_lim
        raise_if_negative(sigma_cp=sigma_cp, sigma_c_lim=sigma_c_lim)

    @property
    def comparison(self) -> bool:
        """Evaluate the comparison σcp ≤ σc,lim.

        Returns
        -------
        bool
            True if σcp ≤ σc,lim, False otherwise.
        """
        return self.sigma_cp <= self.sigma_c_lim

    def __bool__(self) -> bool:
        """Evaluates the comparison, for more information see the __init__ method."""
        return self.comparison

    def __str__(self) -> str:
        """Return the result of the comparison."""
        return self.latex().complete

    def latex(self) -> LatexFormula:
        """Returns LatexFormula object for the comparison."""
        return LatexFormula(
            return_symbol=r"CHECK",
            result="OK" if self.__bool__() else "\\text{Not OK}",
            equation=r"\sigma_{cp} ≤ \sigma_{c,lim}",
            numeric_equation=rf"{self.sigma_cp:.3f} ≤ {self.sigma_c_lim:.3f}",
            comparison_operator_label=r"\rightarrow",
        )



"""Testing formula 5.1 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_5_structural_analysis.formula_5_1 import (
    Form5Dot1Imperfections,
    SubForm5Dot1ReductionFactorLengthOrHeight,
    SubForm5Dot1ReductionFactorNumberOfMembers,
)
from blueprints.validations import NegativeValueError


class TestForm5Dot1Imperfections:
    """Validation for formula 5.1 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        theta_0 = 0.005
        alpha_h = 0.8
        alpha_m = 0.9

        # Object to test
        form_5_1 = Form5Dot1Imperfections(theta_0=theta_0, alpha_h=alpha_h, alpha_m=alpha_m)

        # Expected result, manually calculated
        manually_calculated_result = 0.0036

        assert form_5_1 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_raise_error_when_negative_theta_0_is_given(self) -> None:
        """Test a negative value of Θ0."""
        # Example values
        theta_0 = -0.005
        alpha_h = 0.8
        alpha_m = 0.9

        with pytest.raises(NegativeValueError):
            Form5Dot1Imperfections(theta_0=theta_0, alpha_h=alpha_h, alpha_m=alpha_m)

    def test_raise_error_when_negative_alpha_h_is_given(self) -> None:
        """Test a negative value of αh."""
        # Example values
        theta_0 = 0.005
        alpha_h = -0.8
        alpha_m = 0.9

        with pytest.raises(NegativeValueError):
            Form5Dot1Imperfections(theta_0=theta_0, alpha_h=alpha_h, alpha_m=alpha_m)

    def test_raise_error_when_negative_alpha_m_is_given(self) -> None:
        """Test a negative value of αm."""
        # Example values
        theta_0 = 0.005
        alpha_h = 0.8
        alpha_m = -0.9

        with pytest.raises(NegativeValueError):
            Form5Dot1Imperfections(theta_0=theta_0, alpha_h=alpha_h, alpha_m=alpha_m)

    def test_integration_with_sub_formula_5_1_reduction_factor_length_or_height(self) -> None:
        """Test the integration with sub-formula 5.1 for the reduction factor for length or height, αh."""
        # Example values
        theta_0 = 0.005
        alpha_h = SubForm5Dot1ReductionFactorLengthOrHeight(length=5.3)
        alpha_m = 0.9

        # Object to test
        form_5_1 = Form5Dot1Imperfections(theta_0=theta_0, alpha_h=alpha_h, alpha_m=alpha_m)

        # Expected result, manually calculated
        manually_calculated_result = 0.003909

        assert form_5_1 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_integration_with_sub_formula_5_1_reduction_factor_number_of_members(self) -> None:
        """Test the integration with sub-formula 5.1 for the reduction factor for number of members, αm."""
        # Example values
        theta_0 = 0.005
        alpha_h = 0.8
        alpha_m = SubForm5Dot1ReductionFactorNumberOfMembers(members=3)

        # Object to test
        form_5_1 = Form5Dot1Imperfections(theta_0=theta_0, alpha_h=alpha_h, alpha_m=alpha_m)

        # Expected result, manually calculated
        manually_calculated_result = 0.00326598

        assert form_5_1 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"\alpha_m = \sqrt{0.5 \cdot ( 1 + 1 / m)} = \sqrt{0.5 \cdot ( 1 + 1 / 3.000)} = 0.816",
            ),
            ("short", r"\alpha_m = 0.816"),
        ],
    )
    def test_latex_sub2(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        members = 3

        # Object to test
        form_5_1_sub2_latex = SubForm5Dot1ReductionFactorNumberOfMembers(members).latex()

        actual = {
            "complete": form_5_1_sub2_latex.complete,
            "short": form_5_1_sub2_latex.short,
        }

        assert actual[representation] == expected, f"{representation} representation failed."

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"\alpha_h = \min( \max(2 / \sqrt{l}; 2/3); 1) = \min( \max(2 / \sqrt{5.300}; 2/3); 1) = 0.869",
            ),
            ("short", r"\alpha_h = 0.869"),
        ],
    )
    def test_latex_sub1(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        length = 5.3

        # Object to test
        form_5_1_sub1_latex = SubForm5Dot1ReductionFactorLengthOrHeight(length).latex()

        actual = {
            "complete": form_5_1_sub1_latex.complete,
            "short": form_5_1_sub1_latex.short,
        }

        assert actual[representation] == expected, f"{representation} representation failed."

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"\theta_i = \theta_0 \cdot \alpha_h \cdot \alpha_m = 0.005 \cdot 0.800 \cdot 0.900 = 0.0036",
            ),
            ("short", r"\theta_i = 0.0036"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        theta_0 = 0.005
        alpha_h = 0.8
        alpha_m = 0.9

        # Object to test
        form_5_1_latex = Form5Dot1Imperfections(theta_0=theta_0, alpha_h=alpha_h, alpha_m=alpha_m).latex()

        actual = {
            "complete": form_5_1_latex.complete,
            "short": form_5_1_latex.short,
        }

        assert actual[representation] == expected, f"{representation} representation failed."



"""Testing formula 6.3 of EN 1993-1-1:2005."""

import pytest

from blueprints.codes.eurocode.en_1993_1_1_2005.chapter_6_ultimate_limit_state.formula_6_3 import (
    Form6Dot3MinDeductionAreaStaggeredFastenerHoles,
)
from blueprints.validations import LessOrEqualToZeroError, ListsNotSameLengthError, NegativeValueError


class TestForm6Dot3MinDeductionAreaStaggeredFastenerHoles:
    """Validation for formula 6.3 from EN 1993-1-1:2005."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        t = 10.0
        n = 5.0
        d_0 = 20.0
        s = [50.0, 60.0]
        p = [100.0, 120.0]

        # Object to test
        formula = Form6Dot3MinDeductionAreaStaggeredFastenerHoles(t=t, n=n, d_0=d_0, s=s, p=p)

        # Expected result, manually calculated
        manually_calculated_result = 862.5  # mm^2

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("t", "n", "d_0", "s", "p"),
        [
            (-10.0, 5.0, 20.0, [50.0, 60.0], [100.0, 120.0]),  # t is negative
            (10.0, -5.0, 20.0, [50.0, 60.0], [100.0, 120.0]),  # n is negative
            (10.0, 5.0, -20.0, [50.0, 60.0], [100.0, 120.0]),  # d_0 is negative
            (10.0, 5.0, 20.0, [-50.0, 60.0], [100.0, 120.0]),  # s has a negative value
            (10.0, 5.0, 20.0, [50.0, 60.0], [-100.0, 120.0]),  # p has a negative value
            (0.0, 5.0, 20.0, [50.0, 60.0], [100.0, 120.0]),  # t is zero
            (10.0, 0.0, 20.0, [50.0, 60.0], [100.0, 120.0]),  # n is zero
            (10.0, 5.0, 0.0, [50.0, 60.0], [100.0, 120.0]),  # d_0 is zero
            (10.0, 5.0, 20.0, [0.0, 60.0], [100.0, 120.0]),  # s has a zero value
            (10.0, 5.0, 20.0, [50.0, 60.0], [0.0, 120.0]),  # p has a zero value
            (10.0, 5.0, 20.0, [50.0, 60.0], [100.0, 120.0, 10.0]),  # p has a zero value
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, t: float, n: float, d_0: float, s: list[float], p: list[float]) -> None:
        """Test invalid values."""
        with pytest.raises((NegativeValueError, LessOrEqualToZeroError, ListsNotSameLengthError)):
            Form6Dot3MinDeductionAreaStaggeredFastenerHoles(t=t, n=n, d_0=d_0, s=s, p=p)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"A_{deduction} = t \left( n \cdot d_0 - \sum \frac{s^2}{4 \cdot p} \right) = "
                r"10.000 \left( 5.000 \cdot 20.000 - \left( \frac{50.000^2}{4 \cdot 100.000} + "
                r"\frac{60.000^2}{4 \cdot 120.000} \right) \right) = 862.500 \ mm^2",
            ),
            ("short", r"A_{deduction} = 862.500 \ mm^2"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        t = 10.0
        n = 5.0
        d_0 = 20.0
        s = [50.0, 60.0]
        p = [100.0, 120.0]

        # Object to test
        latex = Form6Dot3MinDeductionAreaStaggeredFastenerHoles(t=t, n=n, d_0=d_0, s=s, p=p).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Formula 7.21 from EN 1992-1-1:2004: Chapter 7 - Serviceability Limit State."""

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula, latex_replace_symbols
from blueprints.type_alias import DIMENSIONLESS, MM3, MM4, MPA, ONE_OVER_MM
from blueprints.validations import raise_if_negative


class Form7Dot21CurvatureDueToShrinkage(Formula):
    r"""Class representing formula 7.21 for the calculation of [$\frac{1}{r_{cs}}$]."""

    label = "7.21"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        epsilon_cs: DIMENSIONLESS,
        es: MPA,
        ec_eff: MPA,
        capital_s: MM3,
        capital_i: MM4,
    ) -> None:
        r"""[$\frac{1}{r_{cs}}$] Calculation of the curvature due to shrinkage [$mm^{-1}$].

        S and i should be calculated for the uncracked condition and the fully cracked condition, the
        final curvature being assessed by use of Expression (7.18).

        EN 1992-1-1:2004 art.7.4.3(6) - Formula (7.21)

        Parameters
        ----------
        epsilon_cs : DIMENSIONLESS
            [$\epsilon_{cs}$] Free shrinkage strain, see 3.1.4.
        es : MPA
            [$E_s$] Modulus of elasticity of the reinforcement [$MPa$].
        ec_eff : MPA
            [$E_{c,eff}$] Effective modulus of elasticity of the concrete [$MPa$].
        s : MM3
            [$S$] First moment of area of the reinforcement about the centroid of the section [$mm^3$].
        i : MM4
            [$I$] Second moment of area of the section [$mm^4$].
        """
        super().__init__()
        self.epsilon_cs = epsilon_cs
        self.es = es
        self.ec_eff = ec_eff
        self.capital_s = capital_s
        self.capital_i = capital_i

    @staticmethod
    def _evaluate(
        epsilon_cs: DIMENSIONLESS,
        es: MPA,
        ec_eff: MPA,
        capital_s: MM3,
        capital_i: MM4,
    ) -> ONE_OVER_MM:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(epsilon_cs=epsilon_cs, es=es, ec_eff=ec_eff, capital_s=capital_s, capital_i=capital_i)

        return epsilon_cs * es / ec_eff * (capital_s / capital_i)

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 7.21."""
        _equation: str = r"\epsilon_{cs} \cdot \frac{E_s}{E_{c,eff}} \cdot \frac{S}{I}"
        _numeric_equation: str = latex_replace_symbols(
            _equation,
            {
                r"\epsilon_{cs}": f"{self.epsilon_cs:.{n + 1}f}",
                r"E_s": f"{self.es:.{n}f}",
                r"E_{c,eff}": f"{self.ec_eff:.{n}f}",
                r"S": f"{self.capital_s:.{n}f}",
                r"I": f"{self.capital_i:.{n}f}",
            },
            False,
        )
        return LatexFormula(
            return_symbol=r"\frac{1}{r_{cs}}",
            result=f"{self:.{n + 3}f}",
            equation=_equation,
            numeric_equation=_numeric_equation,
            comparison_operator_label="=",
            unit="mm^{-1}",
        )



"""Formula 5.46 from EN 1992-1-1:2004: Chapter 5 - Structural Analysis."""

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DIMENSIONLESS, MM, MM2, MM4, MPA, N
from blueprints.validations import raise_if_less_or_equal_to_zero, raise_if_negative


class Form5Dot46Part1TimeDependentForceLosses(Formula):
    r"""Class representing formula 5.46 for the calculation of the time dependent pre- and post-tensioning losses at location
    x under the permanent loads, [$\Delta P_{c+s+r}$].
    """

    label = "5.46"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        a_p: MM2,
        epsilon_cs: DIMENSIONLESS,
        e_p: MPA,
        e_cm: MPA,
        delta_sigma_pr: MPA,
        phi_t_t0: DIMENSIONLESS,
        sigma_c_qp: MPA,
        a_c: MM2,
        i_c: MM4,
        z_cp: MM,
    ) -> None:
        r"""[$\Delta P_{c+s+r}$] Time dependent pre- and post-tensioning losses at location x under the permanent loads [$N$].

        EN 1992-1-1:2004 art.5.10.6(2) - Formula (5.46)

        Parameters
        ----------
        a_p : MM2
            [$A_p$] Area of all the prestressing tendons at the location x [$mm^2$].
        epsilon_cs : DIMENSIONLESS
            [$\epsilon_{cs}$] The estimated shrinkage strain according to 3.1.4(6) in absolute value [$-$].
        e_p : MPA
            [$E_p$] Modulus of elasticity for the prestressing steel, see 3.3.3 (9) [$MPa$].
        e_cm : MPA
            [$E_{cm}$] Modulus of elasticity for the concrete (Table 3.1) [$MPa$].
        delta_sigma_pr : MPA
            [$\Delta \sigma_{pr}$] is the absolute value of the variation of stress in the tendons at location x, at
            time t, due to the relaxation of the prestressing steel. It is determined for a stress of
            [$\sigma_p = \sigma_p(G+P_{m0}+\Psi_2 Q)$]where [$\sigma_p = \sigma_p(G+P_{m0}+\Psi_2 Q)$] is the initial
            stress in the tendons due to initial prestress and quasi-permanent actions. [$MPa$].
        phi_t_t0 : DIMENSIONLESS
            [$\phi(t, t_0)$] Creep coefficient at a time t and load application at time t0 [$-$].
        sigma_c_qp : MPA
            [$\sigma_{c,QP}$] stress in the concrete adjacent to the tendons, due to self-weight and
            initial prestress and other quasi-permanent actions where relevant. [$MPa$].
        a_c : MM2
            [$A_c$] Area of concrete section [$mm^2$].
        i_c : MM4
            [$I_c$] Second moment of area of concrete section [$mm^4$].
        z_cp : MM
            [$z_{cp}$] Distance between the centre of gravity of the concrete section and the tendons [$mm$].
        """
        super().__init__()
        self.a_p = a_p
        self.epsilon_cs = epsilon_cs
        self.e_p = e_p
        self.e_cm = e_cm
        self.delta_sigma_pr = delta_sigma_pr
        self.phi_t_t0 = phi_t_t0
        self.sigma_c_qp = sigma_c_qp
        self.a_c = a_c
        self.i_c = i_c
        self.z_cp = z_cp

    @staticmethod
    def _evaluate(
        a_p: MM2,
        epsilon_cs: DIMENSIONLESS,
        e_p: MPA,
        e_cm: MPA,
        delta_sigma_pr: MPA,
        phi_t_t0: DIMENSIONLESS,
        sigma_c_qp: MPA,
        a_c: MM2,
        i_c: MM4,
        z_cp: MM,
    ) -> N:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(a_p=a_p)

        return a_p * Form5Dot46Part2TimeDependentStressLosses(a_p, epsilon_cs, e_p, e_cm, delta_sigma_pr, phi_t_t0, sigma_c_qp, a_c, i_c, z_cp)

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 5.46."""
        return LatexFormula(
            return_symbol=r"\Delta P_{c+s+r}",
            result=f"{self:.{n}f}",
            equation=r"A_p \cdot \frac{\epsilon_{cs} \cdot E_p + 0.8 \cdot \Delta \sigma_{pr} + \frac{E_p}{E_{cm}} \cdot "
            r"\phi(t, t_0) \cdot \sigma_{c,QP}}{1 + \frac{E_p}{E_{cm}} \cdot \frac{A_p}{A_c} \cdot \left(1 + \frac{A_c}{I_c} "
            r"\cdot z_{cp}^2\right) \cdot \left(1 + 0.8 \cdot \phi(t, t_0)\right)}",
            numeric_equation=rf"{self.a_p:.{n}f} \cdot \frac{{{self.epsilon_cs:.6f} \cdot {self.e_p:.{n}f} + 0.800"
            rf" \cdot {self.delta_sigma_pr:.{n}f} + \frac{{{self.e_p:.{n}f}}}{{{self.e_cm:.{n}f}}} \cdot {self.phi_t_t0:.{n}f} "
            rf"\cdot {self.sigma_c_qp:.{n}f}}}{{1 + \frac{{{self.e_p:.{n}f}}}{{{self.e_cm:.{n}f}}} "
            rf"\cdot \frac{{{self.a_p:.{n}f}}}{{{self.a_c:.{n}f}}} "
            rf"\cdot \left(1 + \frac{{{self.a_c:.{n}f}}}{{{self.i_c:.{n}f}}} \cdot {self.z_cp:.{n}f}^2\right) \cdot \left(1 + 0.800 "
            rf"\cdot {self.phi_t_t0:.{n}f}\right)}}",
            comparison_operator_label="=",
            unit="N",
        )


class Form5Dot46Part2TimeDependentStressLosses(Formula):
    r"""Class representing formula 5.46 for the calculation of the time dependent pre- and post-tensioning losses at location
    x under the permanent loads, [$\Delta \sigma_{p,c+s+r}$].
    """

    label = "5.46"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        a_p: MM2,
        epsilon_cs: DIMENSIONLESS,
        e_p: MPA,
        e_cm: MPA,
        delta_sigma_pr: MPA,
        phi_t_t0: DIMENSIONLESS,
        sigma_c_qp: MPA,
        a_c: MM2,
        i_c: MM4,
        z_cp: MM,
    ) -> None:
        r"""[$\Delta \sigma_{p,c+s+r}$] Time dependent pre- and post-tensioning stress losses at
        location x under the permanent loads [$MPa$].

        EN 1992-1-1:2004 art.5.10.6(2) - Formula (5.46)

        Parameters
        ----------
        a_p : MM2
            [$A_p$] Area of all the prestressing tendons at the location x [$mm^2$].
        epsilon_cs : DIMENSIONLESS
            [$\epsilon_{cs}$] The estimated shrinkage strain according to 3.1.4(6) in absolute value [$-$].
        e_p : MPA
            [$E_p$] Modulus of elasticity for the prestressing steel, see 3.3.3 (9) [$MPa$].
        e_cm : MPA
            [$E_{cm}$] Modulus of elasticity for the concrete (Table 3.1) [$MPa$].
        delta_sigma_pr : MPA
            [$\Delta \sigma_{pr}$] is the absolute value of the variation of stress in the tendons at location x, at
            time t, due to the relaxation of the prestressing steel. It is determined for a stress of
            [$\sigma_p = \sigma_p(G+P_{m0}+\Psi_2 Q)$]where [$\sigma_p = \sigma_p(G+P_{m0}+\Psi_2 Q)$] is the initial
            stress in the tendons due to initial prestress and quasi-permanent actions. [$MPa$].
        phi_t_t0 : DIMENSIONLESS
            [$\phi(t, t_0)$] Creep coefficient at a time t and load application at time t0 [$-$].
        sigma_c_qp : MPA
            [$\sigma_{c,QP}$] stress in the concrete adjacent to the tendons, due to self-weight and
            initial prestress and other quasi-permanent actions where relevant. [$MPa$].
        a_c : MM2
            [$A_c$] Area of concrete section [$mm^2$].
        i_c : MM4
            [$I_c$] Second moment of area of concrete section [$mm^4$].
        z_cp : MM
            [$z_{cp}$] Distance between the centre of gravity of the concrete section and the tendons [$mm$].
        """
        super().__init__()
        self.a_p = a_p
        self.epsilon_cs = epsilon_cs
        self.e_p = e_p
        self.e_cm = e_cm
        self.delta_sigma_pr = delta_sigma_pr
        self.phi_t_t0 = phi_t_t0
        self.sigma_c_qp = sigma_c_qp
        self.a_c = a_c
        self.i_c = i_c
        self.z_cp = z_cp

    @staticmethod
    def _evaluate(
        a_p: MM2,
        epsilon_cs: DIMENSIONLESS,
        e_p: MPA,
        e_cm: MPA,
        delta_sigma_pr: MPA,
        phi_t_t0: DIMENSIONLESS,
        sigma_c_qp: MPA,
        a_c: MM2,
        i_c: MM4,
        z_cp: MM,
    ) -> MPA:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(
            epsilon_cs=epsilon_cs,
            delta_sigma_pr=delta_sigma_pr,
            phi_t_t0=phi_t_t0,
            sigma_c_qp=sigma_c_qp,
            z_cp=z_cp,
        )
        raise_if_less_or_equal_to_zero(a_p=a_p, e_p=e_p, e_cm=e_cm, a_c=a_c, i_c=i_c)

        numerator = epsilon_cs * e_p + 0.8 * delta_sigma_pr + (e_p / e_cm) * phi_t_t0 * sigma_c_qp
        denominator = 1 + (e_p / e_cm) * (a_p / a_c) * (1 + (a_c / i_c) * z_cp**2) * (1 + 0.8 * phi_t_t0)
        return numerator / denominator

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 5.46 for stress losses."""
        return LatexFormula(
            return_symbol=r"\Delta \sigma_{p,c+s+r}",
            result=f"{self:.{n}f}",
            equation=r"\frac{\epsilon_{cs} \cdot E_p + 0.8 \cdot \Delta \sigma_{pr} + \frac{E_p}{E_{cm}} \cdot "
            r"\phi(t, t_0) \cdot \sigma_{c,QP}}{1 + \frac{E_p}{E_{cm}} \cdot \frac{A_p}{A_c} \cdot \left(1 + \frac{A_c}{I_c} "
            r"\cdot z_{cp}^2\right) \cdot \left(1 + 0.8 \cdot \phi(t, t_0)\right)}",
            numeric_equation=rf"\frac{{{self.epsilon_cs:.6f} \cdot {self.e_p:.{n}f} + 0.800"
            rf" \cdot {self.delta_sigma_pr:.{n}f} + \frac{{{self.e_p:.{n}f}}}{{{self.e_cm:.{n}f}}} \cdot {self.phi_t_t0:.{n}f} "
            rf"\cdot {self.sigma_c_qp:.{n}f}}}{{1 + \frac{{{self.e_p:.{n}f}}}{{{self.e_cm:.{n}f}}} "
            rf"\cdot \frac{{{self.a_p:.{n}f}}}{{{self.a_c:.{n}f}}} "
            rf"\cdot \left(1 + \frac{{{self.a_c:.{n}f}}}{{{self.i_c:.{n}f}}} \cdot {self.z_cp:.{n}f}^2\right) \cdot \left(1 + 0.800 "
            rf"\cdot {self.phi_t_t0:.{n}f}\right)}}",
            comparison_operator_label="=",
            unit="MPa",
        )



"""Tests for cross-section shapes."""

import pytest
from shapely.geometry import Polygon

from blueprints.structural_sections.cross_section_triangle import RightAngledTriangularCrossSection


class TestRightAngledTriangularCrossSection:
    """Tests for the RightAngledTriangularCrossSection class."""

    def test_area(self, triangular_cross_section: RightAngledTriangularCrossSection) -> None:
        """Test the area property of the RightAngledTriangularCrossSection class."""
        assert triangular_cross_section.area == pytest.approx(expected=10000.0, rel=1e-6)

    def test_polygon(self, triangular_cross_section: RightAngledTriangularCrossSection) -> None:
        """Test the polygon property of the RightAngledTriangularCrossSection class."""
        assert isinstance(triangular_cross_section.polygon, Polygon)

    def test_geometry(self, triangular_cross_section: RightAngledTriangularCrossSection) -> None:
        """Test the geometry property of the RightAngledTriangularCrossSection class."""
        geometry = triangular_cross_section.geometry()
        assert geometry is not None

    def test_invalid_base(self) -> None:
        """Test initialization with an invalid base value."""
        with pytest.raises(ValueError, match="Base must be a positive value"):
            RightAngledTriangularCrossSection(name="InvalidBase", base=-10.0, height=200.0)

    def test_invalid_height(self) -> None:
        """Test initialization with an invalid height value."""
        with pytest.raises(ValueError, match="Height must be a positive value"):
            RightAngledTriangularCrossSection(name="InvalidHeight", base=100.0, height=-20.0)

    def test_mirrored_polygon(self) -> None:
        """Test the geometry property when the triangle is mirrored."""
        mirrored_triangle = RightAngledTriangularCrossSection(
            name="MirroredTriangle", base=100.0, height=200.0, mirrored_horizontally=True, mirrored_vertically=True
        )
        polygon = mirrored_triangle.polygon
        assert polygon.is_valid
        assert len(polygon.exterior.coords) == 4
        assert (polygon.exterior.coords[1][0], polygon.exterior.coords[1][1]) == (-100.0, 0)
        assert (polygon.exterior.coords[2][0], polygon.exterior.coords[2][1]) == (0, -200.0)



"""Module for the abstract base class Formula."""

from abc import ABC, abstractmethod

from blueprints.codes.latex_formula import LatexFormula


class Formula(float, ABC):
    """Abstract base class for formulas used in the codes."""

    def __new__(cls, *args, **kwargs) -> "Formula":
        """Method for creating a new instance of the class."""
        result = cls._evaluate(*args, **kwargs)
        instance = float.__new__(cls, result)
        instance._initialized = False  # noqa: SLF001
        return instance

    def __init__(self, *args, **kwargs) -> None:
        """Method for initializing a new instance of the class."""
        super().__init__(*args, **kwargs)
        self._initialized = True

    def __setattr__(self, name: str, value: str | float) -> None:
        """Override the __setattr__ method to prevent modifications after initialization.

        Parameters
        ----------
        name : str
            The name of the attribute.
        value : str | float
            The value to be assigned to the attribute.
        """
        if getattr(self, "_initialized", False) and name in self.__dict__:
            raise AttributeError(f"Attribute '{name}' of '{type(self).__name__}' object is read-only and cannot be modified after initialization.")
        super().__setattr__(name, value)

    @property
    @abstractmethod
    def label(self) -> str:
        """Property for the formula label.

        For example, "5.2" for formula 5.2.

        Returns
        -------
        str
            The label/number associated with the formula.
            This is an abstract method and must be implemented in all subclasses.
        """

    @property
    @abstractmethod
    def source_document(self) -> str:
        """Property for the source document.

        For example, "EN 1992-1-1:2004"
        Try to use the official and complete name of the document including publishing year, if possible.

        Returns
        -------
        str
            The reference to the document where the formula originates.
            This is an abstract method and must be implemented in all subclasses.
        """

    @property
    def detailed_result(self) -> dict:
        """Property for providing the detailed result of the formula.

        Returns
        -------
        dict
            The detailed result of the formula.
            Keys are strings representing the name of the partial or intermediate result.
            Values types will depend on the specific implementation, but must be a serializable type.
        """
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def _evaluate(*args, **kwargs) -> float | bool:
        """Abstract method for the logic of the formula.

        Returns
        -------
        float | bool
            The result of the formula.
            This is an abstract method and must be implemented in all subclasses.
        """

    @abstractmethod
    def latex(self, n: int = 3) -> LatexFormula:
        """Abstract method for the latex representation of the formula.

        Parameters
        ----------
        n : int, optional
            The number of decimal places to round the result to.

        Returns
        -------
        LatexFormula
            The latex representation of the formula.
            This is an abstract method and must be implemented in all subclasses.
        """


class ComparisonFormula(Formula):
    """Base class for comparison formulas used in the codes."""

    def __new__(cls, *args, **kwargs) -> "ComparisonFormula":
        """Method for creating a new instance of the class."""
        lhs = cls._evaluate_lhs(*args, **kwargs)
        rhs = cls._evaluate_rhs(*args, **kwargs)
        result = cls._evaluate(*args, **kwargs)
        instance = float.__new__(cls, result)
        instance._lhs = lhs  # noqa: SLF001
        instance._rhs = rhs  # noqa: SLF001
        instance._initialized = False  # noqa: SLF001
        return instance

    @staticmethod
    @abstractmethod
    def _evaluate_lhs(*args, **kwargs) -> float:
        """Abstract method for the logic of the left-hand side of the comparison formula.

        Returns
        -------
        float
            The left-hand side value of the comparison.
        """

    @staticmethod
    @abstractmethod
    def _evaluate_rhs(*args, **kwargs) -> float:
        """Abstract method for the logic of the right-hand side of the comparison formula.

        Returns
        -------
        float
            The right-hand side value of the comparison.
        """

    @property
    def lhs(self) -> float:
        """Property for getting the left-hand side of the comparison.

        Returns
        -------
        float
            The left-hand side value of the comparison.
        """
        return self._lhs  # type: ignore[attr-defined]

    @property
    def rhs(self) -> float:
        """Property for getting the right-hand side of the comparison.

        Returns
        -------
        float
            The right-hand side value of the comparison.
        """
        return self._rhs  # type: ignore[attr-defined]

    @property
    @abstractmethod
    def unity_check(self) -> float:
        """Property to present the unity check of the formula.

        Returns
        -------
        float
            The unity check.
        """



"""Testing sub-formula 1 from 8.15 from EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_8_detailing_of_reinforcement_and_prestressing_tendons.formula_8_15 import (
    SubForm8Dot15EtaP1,
)


class TestSubForm8Dot15EtaP1:
    """Validation for sub-formula 1 from 8.15 from EN 1992-1-1:2004."""

    @pytest.mark.parametrize(
        ("type_of_wire", "expected"),
        [
            ("indented", 2.7),
            ("3_7_wire_strands", 3.2),
        ],
    )
    def test_evaluation(self, type_of_wire: str, expected: float) -> None:
        """Test the evaluation of the result."""
        # example values
        sub_form_8_15 = SubForm8Dot15EtaP1(
            type_of_wire=type_of_wire,
        )
        assert sub_form_8_15 == pytest.approx(expected=expected, rel=1e-4)

    def test_invalid_type_of_wire(self) -> None:
        """Test the evaluation of the result."""
        # example values
        type_of_wire = "invalid"
        with pytest.raises(ValueError):
            SubForm8Dot15EtaP1(
                type_of_wire=type_of_wire,
            )

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                (r"\eta_{p1} \rightarrow type\;of\;wire \rightarrow indented \rightarrow 2.70"),
            ),
            ("short", r"\eta_{p1} \rightarrow 2.70"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # example values
        type_of_wire = "indented"

        # Object to test
        form_8_15_p1_latex = SubForm8Dot15EtaP1(type_of_wire=type_of_wire).latex()

        actual = {"complete": form_8_15_p1_latex.complete, "short": form_8_15_p1_latex.short}

        assert actual[representation] == expected, f"{representation} representation failed."



"""Shape of a circular cornered section."""

from dataclasses import dataclass

import numpy as np
from sectionproperties.pre import Geometry
from shapely.geometry import Polygon

from blueprints.structural_sections._cross_section import CrossSection
from blueprints.type_alias import MM
from blueprints.validations import raise_if_negative


@dataclass(frozen=True)
class CircularCorneredCrossSection(CrossSection):
    """
    Class to represent a square cross-section with a quarter circle cutout for geometric calculations, named as a circular cornered section.

        .---- outer arc
        ∨
    . . .+-----------------------+
    .  ⁄                         |
    .⁄                           |<-- thickness_vertical
    +                            |
    |                        _ _ |<-- inner arc
    |                      /
    |                    /
    |                   |
    +-------------------+<-- thickness_horizontal

    Parameters
    ----------
    thickness_vertical : MM
        Thickness of the vertical section
    thickness_horizontal : MM
        Thickness of the horizontal section
    inner_radius : MM
        Inner radius of the corner
    outer_radius : MM
        Outer radius of the corner
    corner_direction : int
        ↰ = 0, ↱ = 1, ↳ = 2, ↲ = 3
    x : MM
        x-coordinate of the center of the inner_radius (default 0)
    y : MM
        y-coordinate of the center of the inner_radius (default 0)
    name : str
        Name of the cross-section (default "Corner")
    """

    thickness_vertical: MM
    thickness_horizontal: MM
    inner_radius: MM
    outer_radius: MM
    x: MM = 0
    y: MM = 0
    corner_direction: int = 0  # 0 = ↰, 1 = ↱, 2 = ↳, 3 = ↲
    name: str = "Corner"

    def __post_init__(self) -> None:
        """Validate input parameters after initialization."""
        raise_if_negative(
            thickness_vertical=self.thickness_vertical,
            thickness_horizontal=self.thickness_horizontal,
            inner_radius=self.inner_radius,
            outer_radius=self.outer_radius,
        )
        if self.outer_radius > self.inner_radius + min(self.thickness_vertical, self.thickness_horizontal):
            raise ValueError(
                f"Outer radius {self.outer_radius} must be smaller than or equal to inner radius {self.inner_radius} "
                f"plus the thickness {min(self.thickness_vertical, self.thickness_horizontal)}"
            )
        if self.corner_direction not in (0, 1, 2, 3):
            raise ValueError(f"corner_direction must be one of 0, 1, 2, or 3, got {self.corner_direction}")

    @property
    def width_rectangle(self) -> MM:
        """Width of the rectangle part of the corner cross-section [mm]."""
        return self.thickness_horizontal + self.inner_radius

    @property
    def height_rectangle(self) -> MM:
        """Height of the rectangle part of the corner cross-section [mm]."""
        return self.thickness_vertical + self.inner_radius

    @property
    def polygon(self) -> Polygon:
        """Shapely Polygon representing the corner cross-section."""
        lr = (self.x + self.width_rectangle, self.y)
        ul = (self.x, self.y + self.height_rectangle)

        n = 16

        # Outer arc (from vertical to horizontal)
        theta_outer = np.linspace(0, np.pi / 2, n)
        outer_arc = np.column_stack(
            (
                self.x + self.width_rectangle - self.outer_radius + self.outer_radius * np.cos(theta_outer),
                self.y + self.height_rectangle - self.outer_radius + self.outer_radius * np.sin(theta_outer),
            )
        )

        # Inner arc (from horizontal to vertical, reversed)
        theta_inner = np.linspace(0, np.pi / 2, n)
        inner_arc = np.column_stack(
            (
                self.x + self.inner_radius * np.cos(theta_inner),
                self.y + self.inner_radius * np.sin(theta_inner),
            )
        )[::-1]

        # Combine points
        points = np.vstack([lr, outer_arc, ul, inner_arc])

        # Remove consecutive duplicate points
        diff = np.diff(points, axis=0)
        mask = np.any(diff != 0, axis=1)
        mask = np.insert(mask, 0, True)
        points = points[mask]

        # Create transformation matrices for flipping
        flip_x = np.array([[-1, 0], [0, 1]])
        flip_y = np.array([[1, 0], [0, -1]])

        # Center points around (self.x, self.y)
        points_centered = points - np.array([self.x, self.y])

        # Apply flips based on corner_direction
        if self.corner_direction in (1, 2):
            points_centered = points_centered @ flip_x
        if self.corner_direction in (2, 3):
            points_centered = points_centered @ flip_y

        # Shift points back
        points = points_centered + np.array([self.x, self.y])

        points = np.array([tuple(pt) for pt in points])

        return Polygon(points)

    def geometry(self, mesh_size: MM | None = None) -> Geometry:
        """
        Return the geometry of the RHSCF corner cross-section.

        Parameters
        ----------
        mesh_size : MM | None
            Maximum mesh element area to be used within the Geometry-object finite-element mesh. If not provided, a default value
            will be used with maximum of 2 mm or halve the smallest thickness.
        """
        if mesh_size is None:
            minimum_mesh_size = 2.0
            mesh_length = max(min(self.thickness_vertical, self.thickness_horizontal) / 2, minimum_mesh_size)
            mesh_size = mesh_length**2

        geom = Geometry(geom=self.polygon)
        geom.create_mesh(mesh_sizes=mesh_size)
        return geom



"""Formula A.1 from EN 1993-1-9:2005: Annex A - Determination of fatigue load parameters and verification formats."""

from blueprints.codes.eurocode.en_1993_1_9_2005 import EN_1993_1_9_2005
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DIMENSIONLESS
from blueprints.validations import raise_if_negative


class FormADot2CriteriaBasedOnDamageAccumulation(Formula):
    """Class representing formula A.2 for the calculation of the damage during the design life."""

    label = "A.2"
    source_document = EN_1993_1_9_2005

    def __init__(
        self,
        d_d: DIMENSIONLESS,
    ) -> None:
        r"""[$CHECK$] Criteria met, based on damage accumulation.

        EN 1993-1-9:2005 art.A.5 - Formula (A.1)

        Parameters
        ----------
        d_d : DIMENSIONLESS
            [$D_d$] The damage during the design life [$-$].

        Returns
        -------
        None
        """
        super().__init__()
        self.d_d = d_d

    @staticmethod
    def _evaluate(
        d_d: DIMENSIONLESS,
    ) -> bool:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(d_d=d_d)
        return d_d <= 1.0

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula A.2."""
        return LatexFormula(
            return_symbol=r"CHECK",
            result="OK" if self <= 1 else r"NOT\;OK",
            equation=r"D_d \leq 1.0",
            numeric_equation=rf"{self.d_d:.{n}f} \leq 1.0",
            comparison_operator_label=r"\rightarrow",
        )



"""Testing formula 6.78/6.79 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_6_ultimate_limit_state.formula_6_78_79 import Form6Dot78And79FatigueResistance
from blueprints.validations import NegativeValueError


class TestForm6Dot78And79FatigueResistance:
    """Validation for formula 6.78/6.79 from EN 1992-1-1:2004."""

    def test_evaluation_78(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        v_ed_max = 100.0
        v_ed_min = 50.0
        v_rd_c = 200.0
        f_ck = 30.0

        # Object to test
        formula = Form6Dot78And79FatigueResistance(v_ed_max=v_ed_max, v_ed_min=v_ed_min, v_rd_c=v_rd_c, f_ck=f_ck)

        # Expected result, manually calculated
        expected_result = True

        assert formula == expected_result

    def test_evaluation_79(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        v_ed_max = 100.0
        v_ed_min = -50.0
        v_rd_c = 200.0
        f_ck = 30.0

        # Object to test
        formula = Form6Dot78And79FatigueResistance(v_ed_max=v_ed_max, v_ed_min=v_ed_min, v_rd_c=v_rd_c, f_ck=f_ck)

        # Expected result, manually calculated
        expected_result = False

        assert formula == expected_result

    @pytest.mark.parametrize(
        ("v_ed_max", "v_ed_min", "v_rd_c", "f_ck"),
        [
            (100.0, 50.0, -200.0, 30.0),  # v_rd_c is negative
            (100.0, 50.0, 200.0, -30.0),  # f_ck is negative
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, v_ed_max: float, v_ed_min: float, v_rd_c: float, f_ck: float) -> None:
        """Test invalid values."""
        with pytest.raises(NegativeValueError):
            Form6Dot78And79FatigueResistance(v_ed_max=v_ed_max, v_ed_min=v_ed_min, v_rd_c=v_rd_c, f_ck=f_ck)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"CHECK \to \begin{cases} \frac{\left|V_{Ed,max}\right|}{\left|V_{Rd,c}\right|} \leq "
                r"\min\left(0.5 + 0.45 \cdot \frac{\left|V_{Ed,min}\right|}{\left|V_{Rd,c}\right|}, "
                r"\begin{cases} 0.9 & \text{if } f_{ck} \le 50 \\ 0.8 & \text{if } f_{ck} > 50 \end{cases} \right) "
                r"& \text{if } \frac{V_{Ed,min}}{V_{Ed,max}} \geq 0 \\ \frac{\left|V_{Ed,max}\right|}"
                r"{\left|V_{Rd,c}\right|} \leq 0.5 - \frac{\left|V_{Ed,min}\right|}{\left|V_{Rd,c}\right|} & \text{if } "
                r"\frac{V_{Ed,min}}{V_{Ed,max}} < 0 \end{cases} \to "
                r"\begin{cases} \frac{\left|100.000\right|}{\left|200.000\right|} \leq \min\left(0.5 + "
                r"0.45 \cdot \frac{\left|50.000\right|}{\left|200.000\right|}, \begin{cases} 0.9 & \text{if } 30.000 "
                r"\le 50 \\ 0.8 & \text{if } 30.000 > 50 \end{cases} \right) & \text{if } \frac{50.000}{100.000} \geq 0 "
                r"\\ \frac{\left|100.000\right|}{\left|200.000\right|} \leq 0.5 - \frac{\left|50.000\right|}{\left|200.000\right|} "
                r"& \text{if } \frac{50.000}{100.000} < 0 \end{cases} \to OK",
            ),
            ("short", r"CHECK \to OK"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        v_ed_max = 100.0
        v_ed_min = 50.0
        v_rd_c = 200.0
        f_ck = 30.0

        # Object to test
        latex = Form6Dot78And79FatigueResistance(v_ed_max=v_ed_max, v_ed_min=v_ed_min, v_rd_c=v_rd_c, f_ck=f_ck).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Testing formula 7.13 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_7_serviceability_limit_state.formula_7_13 import Form7Dot13CoefficientK2
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestForm7Dot13CoefficientK2:
    """Validation for formula 7.13 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        epsilon_1 = 0.002
        epsilon_2 = 0.001

        # Object to test
        formula = Form7Dot13CoefficientK2(epsilon_1=epsilon_1, epsilon_2=epsilon_2)

        # Expected result, manually calculated
        manually_calculated_result = 0.75  # dimensionless

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("epsilon_1", "epsilon_2"),
        [
            (0, 0.001),  # epsilon_1 is zero
            (-0.002, 0.001),  # epsilon_1 is negative
            (0.002, -0.001),  # epsilon_2 is negative
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, epsilon_1: float, epsilon_2: float) -> None:
        """Test invalid values."""
        with pytest.raises((LessOrEqualToZeroError, NegativeValueError)):
            Form7Dot13CoefficientK2(epsilon_1=epsilon_1, epsilon_2=epsilon_2)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"k_2 = \frac{\epsilon_1 + \epsilon_2}{2 \cdot \epsilon_1} = \frac{0.002 + 0.001}{2 \cdot 0.002} = 0.750 \ -",
            ),
            ("short", r"k_2 = 0.750 \ -"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        epsilon_1 = 0.002
        epsilon_2 = 0.001

        # Object to test
        latex = Form7Dot13CoefficientK2(epsilon_1=epsilon_1, epsilon_2=epsilon_2).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Formula 6.25 from EN 1993-1-1:2005: Chapter 6 - Ultimate Limit State."""

from blueprints.codes.eurocode.en_1993_1_1_2005 import EN_1993_1_1_2005
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula, latex_replace_symbols
from blueprints.type_alias import N
from blueprints.validations import raise_if_less_or_equal_to_zero, raise_if_negative


class Form6Dot25CheckCombinedShearForceAndTorsionalMoment(Formula):
    r"""Class representing formula 6.25 for combined shear force and torsional moment."""

    label = "6.25"
    source_document = EN_1993_1_1_2005

    def __init__(
        self,
        v_ed: N,
        v_pl_t_rd: N,
    ) -> None:
        r"""Check the combined shear force and torsional moment.

        EN 1993-1-1:2005 art.6.2.7(9) - Formula (6.25)

        Parameters
        ----------
        v_ed : N
            [$V_{Ed}$] Design shear force [$N$].
        v_pl_t_rd : N
            [$V_{pl,T,Rd}$] Plastic shear resistance accounting for torsional effects, derived from equation 6.26, 6.27 or 6.28 [$N$].
        """
        super().__init__()
        self.v_ed = v_ed
        self.v_pl_t_rd = v_pl_t_rd

    @staticmethod
    def _evaluate(
        v_ed: N,
        v_pl_t_rd: N,
    ) -> bool:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_less_or_equal_to_zero(v_pl_t_rd=v_pl_t_rd)
        raise_if_negative(v_ed=v_ed)

        return v_ed / v_pl_t_rd <= 1.0

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 6.25."""
        _equation: str = r"\left( \frac{V_{Ed}}{V_{pl,T,Rd}} \leq 1.0 \right)"
        _numeric_equation: str = latex_replace_symbols(
            _equation,
            {
                "V_{Ed}": f"{self.v_ed:.{n}f}",
                "V_{pl,T,Rd}": f"{self.v_pl_t_rd:.{n}f}",
            },
            False,
        )
        return LatexFormula(
            return_symbol=r"CHECK",
            result="OK" if self.__bool__() else "\\text{Not OK}",
            equation=_equation,
            numeric_equation=_numeric_equation,
            comparison_operator_label="\\to",
            unit="",
        )



"""Tests for cross-section shapes."""

import pytest
from sectionproperties.analysis import Section

from blueprints.structural_sections.cross_section_annular_sector import AnnularSectorCrossSection


class TestAnnularSectorCrossSection:
    """Tests for the AnnularSectorCrossSection class."""

    def test_width(self, annular_sector_cross_section: AnnularSectorCrossSection) -> None:
        """Test the width property of the AnnularSectorCrossSection class."""
        expected_width = 110.0
        assert annular_sector_cross_section.width == pytest.approx(expected=expected_width, rel=1e-6)

    def test_height(self, annular_sector_cross_section: AnnularSectorCrossSection) -> None:
        """Test the height property of the AnnularSectorCrossSection class."""
        expected_height = 110.0
        assert annular_sector_cross_section.height == pytest.approx(expected=expected_height, rel=1e-6)

    def test_polygon(self, annular_sector_cross_section: AnnularSectorCrossSection) -> None:
        """Test the polygon property of the AnnularSectorCrossSection class."""
        assert annular_sector_cross_section.polygon.is_valid
        assert len(annular_sector_cross_section.polygon.exterior.coords) > 0

    def test_section(self, annular_sector_cross_section: AnnularSectorCrossSection) -> None:
        """Test the section object of the AnnularSectorCrossSection class."""
        section = annular_sector_cross_section.section()
        assert isinstance(section, Section)

    def test_geometry(self, annular_sector_cross_section: AnnularSectorCrossSection) -> None:
        """Test the geometry property of the AnnularSectorCrossSection class."""
        geometry = annular_sector_cross_section.geometry()
        assert geometry is not None

    def test_invalid_radius(self) -> None:
        """Test initialization with an invalid radius value."""
        with pytest.raises(ValueError, match="Radius must be zero or positive"):
            AnnularSectorCrossSection(
                name="InvalidRadius",
                inner_radius=-10.0,
                thickness=20.0,
                start_angle=0.0,
                end_angle=90.0,
                x=0.0,
                y=0.0,
            )

    def test_invalid_thickness(self) -> None:
        """Test initialization with an invalid thickness value."""
        with pytest.raises(ValueError, match="Thickness must be a positive value"):
            AnnularSectorCrossSection(
                name="InvalidThickness",
                inner_radius=100.0,
                thickness=-20.0,
                start_angle=0.0,
                end_angle=90.0,
                x=0.0,
                y=0.0,
            )

    def test_invalid_end_angle_smaller_than_start_angle(self) -> None:
        """Test initialization with an invalid end angle value."""
        with pytest.raises(ValueError):
            AnnularSectorCrossSection(
                name="InvalidEndAngle",
                inner_radius=100.0,
                thickness=20.0,
                start_angle=90.0,
                end_angle=0.0,
                x=0.0,
                y=0.0,
            )

    @pytest.mark.parametrize(
        ("start_angle", "end_angle"),
        [
            (0, 360),
            (0, 720),
            (-50, 360),
            (0, 400),
        ],
    )
    def test_total_angle_greater_or_equal_than_360(self, start_angle: float, end_angle: float) -> None:
        """Test initialization with an invalid end angle value."""
        with pytest.raises(ValueError):
            AnnularSectorCrossSection(
                name="Invalid total angle, greater than 360 degrees",
                inner_radius=100.0,
                thickness=20.0,
                start_angle=start_angle,
                end_angle=end_angle,
                x=0.0,
                y=0.0,
            )

    def test_invalid_start_angle(self) -> None:
        """Test initialization with an invalid end angle value."""
        with pytest.raises(ValueError):
            AnnularSectorCrossSection(
                name="InvalidEndAngle",
                inner_radius=100.0,
                thickness=20.0,
                start_angle=-683,
                end_angle=270,
                x=0.0,
                y=0.0,
            )

    @pytest.mark.parametrize(
        ("start_angle", "end_angle", "expected_area"),
        [
            (0, 90, 3141.5926),
            (0, 180, 3141.5926 * 2),
            (0, 270, 3141.5926 * 3),
            (0, 359.9999999, 3141.5926 * 4),
            (90, 180, 3141.5926),
            (90, 270, 3141.5926 * 2),
            (90, 360, 3141.5926 * 3),
            (-90, 269.999999, 3141.5926 * 4),
            (-180, 90, 3141.5926 * 3),
            (-360, -270, 3141.5926),
        ],
    )
    def test_area_at_different_angles(self, start_angle: float, end_angle: float, expected_area: float) -> None:
        """Test the area property of the AnnularSectorCrossSection class at different angles.
        A full donut (0 to 360 degrees) should have an area of 0.
        a quarter donut (0 to 90 degrees) should have an area of pi * (outer_radius^2 - inner_radius^2) / 4.
        """
        annular_sector = AnnularSectorCrossSection(
            inner_radius=90.0,
            thickness=20.0,
            start_angle=start_angle,
            end_angle=end_angle,
            x=100.0,
            y=250.0,
        )
        assert annular_sector.area == pytest.approx(expected=expected_area, rel=1e-3)



"""Formula 6.26 from EN 1992-1-1:2004: Chapter 6 - Ultimate limit state."""

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import MM2, N_MM, NMM
from blueprints.validations import raise_if_less_or_equal_to_zero, raise_if_negative


class Form6Dot26ShearStressInWall(Formula):
    r"""Class representing formula 6.26 for the calculation of the shear stress in a wall of a section
    subject to a pure torsional moment multiplied with the effective thickness [$\tau_{t,i}t_{ef,i}$].
    """

    label = "6.26"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        t_ed: NMM,
        a_k: MM2,
    ) -> None:
        r"""[$\tau_{t,i}t_{ef,i}$] Shear stress in a wall of a section subject to a pure torsional moment multiplied with the
        effective thickness [$N/m$].

        EN 1992-1-1:2004 art.6.3.2(1) - Formula (6.26)

        Parameters
        ----------
        t_ed : NMM
            [$T_{Ed}$] Applied design torsion [$Nmm$].
        a_k : MM2
            [$A_k$] Area enclosed by the centre-lines of the connecting walls, including inner hollow areas [$mm^2$].
        """
        super().__init__()
        self.t_ed = t_ed
        self.a_k = a_k

    @staticmethod
    def _evaluate(
        t_ed: NMM,
        a_k: MM2,
    ) -> N_MM:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_less_or_equal_to_zero(a_k=a_k)
        raise_if_negative(t_ed=t_ed)

        return t_ed / (2 * a_k)

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 6.26."""
        return LatexFormula(
            return_symbol=r"\tau_{t,i}t_{ef,i}",
            result=f"{self:.{n}f}",
            equation=r"\frac{T_{Ed}}{2 \cdot A_{k}}",
            numeric_equation=rf"\frac{{{self.t_ed:.{n}f}}}{{2 \cdot {self.a_k:.{n}f}}}",
            comparison_operator_label="=",
            unit="N/mm",
        )



"""Formula 6.62 from EN 1992-1-1:2004: Chapter 6 - Ultimate Limit State."""

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula, latex_replace_symbols
from blueprints.type_alias import DIMENSIONLESS, MPA
from blueprints.validations import raise_if_negative


class Form6Dot62DesignValueCompressiveStressResistance(Formula):
    r"""Class representing formula 6.62 for the calculation of [$\sigma_{Rd,max}$]."""

    label = "6.62"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        k_3: DIMENSIONLESS,
        nu_prime: DIMENSIONLESS,
        f_cd: MPA,
    ) -> None:
        r"""[$\sigma_{Rd,max}$] Calculation of [$\sigma_{Rd,max}$].

        EN 1992-1-1:2004 art.6.5.4(4) - Formula (6.62)

        Parameters
        ----------
        k_3 : DIMENSIONLESS
            [$k_3$] Coefficient for the design value of compressive stress resistance [$-$].
            Note: The value of [$k_3$] for use in a Country may be found in its National Annex.
            The recommended value is 0.75.
        nu_prime : DIMENSIONLESS
            [$\nu'$] Reduction factor for the design value of compressive stress resistance [$-$].
        f_cd : MPA
            [$f_{cd}$] Design value of compressive strength [$MPa$].
        """
        super().__init__()
        self.k_3 = k_3
        self.nu_prime = nu_prime
        self.f_cd = f_cd

    @staticmethod
    def _evaluate(
        k_3: DIMENSIONLESS,
        nu_prime: DIMENSIONLESS,
        f_cd: MPA,
    ) -> MPA:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(k_3=k_3, nu_prime=nu_prime, f_cd=f_cd)

        return k_3 * nu_prime * f_cd

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 6.62."""
        _equation: str = r"k_3 \cdot \nu' \cdot f_{cd}"
        _numeric_equation: str = latex_replace_symbols(
            _equation,
            {
                r"k_3": f"{self.k_3:.{n}f}",
                r"\nu'": f"{self.nu_prime:.{n}f}",
                r"f_{cd}": f"{self.f_cd:.{n}f}",
            },
            False,
        )
        return LatexFormula(
            return_symbol=r"\sigma_{Rd,max}",
            result=f"{self:.{n}f}",
            equation=_equation,
            numeric_equation=_numeric_equation,
            comparison_operator_label="=",
            unit="MPa",
        )



"""Formula 6.31 from EN 1993-1-1:2005: Chapter 6 - Ultimate Limit State."""

from blueprints.codes.eurocode.en_1993_1_1_2005 import EN_1993_1_1_2005
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula, latex_replace_symbols
from blueprints.type_alias import NMM
from blueprints.validations import raise_if_negative


class Form6Dot31CheckBendingAndAxialForce(Formula):
    r"""Class representing formula 6.31 for bending and axial force."""

    label = "6.31"
    source_document = EN_1993_1_1_2005

    def __init__(
        self,
        m_ed: NMM,
        m_n_rd: NMM,
    ) -> None:
        r"""Check the bending and axial force criterion.

        EN 1993-1-1:2005 art.6.2.9.1 - Formula (6.31)

        Parameters
        ----------
        m_ed : NMM
            [$M_{Ed}$] Design value of the applied bending moment [$Nmm$].
        m_n_rd : NMM
            [$M_{N,Rd}$] Design plastic moment resistance reduced due to the axial force $N_{Ed}$ [$Nmm$].
        """
        super().__init__()
        self.m_ed = m_ed
        self.m_n_rd = m_n_rd

    @staticmethod
    def _evaluate(
        m_ed: NMM,
        m_n_rd: NMM,
    ) -> bool:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(m_ed=m_ed, m_n_rd=m_n_rd)

        return m_ed <= m_n_rd

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 6.31."""
        _equation: str = r"M_{Ed} \leq M_{N,Rd}"
        _numeric_equation: str = latex_replace_symbols(
            _equation,
            {
                "M_{Ed}": f"{self.m_ed:.{n}f}",
                "M_{N,Rd}": f"{self.m_n_rd:.{n}f}",
            },
            False,
        )
        _numeric_equation_with_units: str = latex_replace_symbols(
            _equation,
            {
                "M_{Ed}": rf"{self.m_ed:.{n}f} \ Nmm",
                "M_{N,Rd}": rf"{self.m_n_rd:.{n}f} \ Nmm",
            },
            False,
        )
        return LatexFormula(
            return_symbol=r"CHECK",
            result="OK" if self.__bool__() else "\\text{Not OK}",
            equation=_equation,
            numeric_equation=_numeric_equation,
            numeric_equation_with_units=_numeric_equation_with_units,
            comparison_operator_label="\\to",
            unit="",
        )



"""Table 4.4N from NEN-EN 1992-1-1:2005+A1:2015+NB:2016+A1:2020: Chapter 4 - Durability and cover to reinforcement."""

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_4_durability_and_cover._base_classes.structural_class import ConcreteStructuralClassBase
from blueprints.codes.eurocode.nen_en_1992_1_1_a1_2020 import NEN_EN_1992_1_1_A1_2020
from blueprints.codes.eurocode.nen_en_1992_1_1_a1_2020.chapter_4_durability_and_cover.table_4_1 import Table4Dot1ExposureClasses
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import MM


class Table4Dot4nMinimumCoverDurabilityReinforcementSteel(Formula):
    r"""Class representing the table 4.4N
    for the calculation of the minimum cover [$c_{min,dur}$] [$mm$] requirements with regard to durability
    for reinforcement steel.
    """

    label = "4.4N"
    source_document = NEN_EN_1992_1_1_A1_2020

    def __init__(
        self,
        exposure_classes: Table4Dot1ExposureClasses,
        structural_class: ConcreteStructuralClassBase,
    ) -> None:
        r"""[$c_{min,dur}$] Calculates the minimum concrete cover with regard to durability [$mm$] for reinforcement steel.

        NEN-EN 1992-1-1:2005+A1:2015+NB:2016+A1:2020 art.4.4.1.2 (5) - Table (4.4N)

        Parameters
        ----------
        exposure_classes: Table4Dot1ExposureClasses
            The exposure classes of the concrete. Use the [$Table4Dot1ExposureClasses$] class. [$-$]
        structural_class: ConcreteStructuralClassBase
            The structural class of the concrete. Use the [$Table4Dot3ConcreteStructuralClass$] class. [$-$]
        """
        super().__init__()
        self.exposure_classes = exposure_classes
        self.structural_class = structural_class

    @staticmethod
    def _evaluate(
        exposure_classes: Table4Dot1ExposureClasses,
        structural_class: ConcreteStructuralClassBase,
    ) -> MM:
        """For more detailed documentation see the class docstring."""
        if not isinstance(structural_class, int):
            raise TypeError(f"Structural class must be (a subclass of) an integer, not {type(structural_class)}.")

        if structural_class < 1 or structural_class > 6:
            raise ValueError("Structural class must be between 1 and 6.")

        if exposure_classes.chloride.value in [
            "XD3",
            "XD2",
        ] or exposure_classes.chloride_seawater.value in ["XS2", "XS3"]:
            calculated_cover = 20 + structural_class * 5
        elif exposure_classes.chloride.value == "XD1" or exposure_classes.chloride_seawater.value == "XS1":
            calculated_cover = 15 + structural_class * 5
        elif exposure_classes.carbonation.value == "XC4":
            calculated_cover = 10 + structural_class * 5
        elif exposure_classes.carbonation.value in ["XC2", "XC3"]:
            calculated_cover = 5 + structural_class * 5
        elif exposure_classes.carbonation.value == "XC1":
            calculated_cover = max(10, ((structural_class - 1) * 5))
        else:
            calculated_cover = max(10, ((structural_class - 2) * 5))

        return calculated_cover

    def latex(self, n: int = 0) -> LatexFormula:
        """Returns LatexFormula object for table 4.4N."""
        return LatexFormula(
            return_symbol=r"c_{min,dur}",
            result=f"{self:.{n}f}",
            equation=rf"structural class {self.structural_class} \& exposure classes ({self.exposure_classes})",
            comparison_operator_label="=",
        )



"""RHS- and SHS-Profile steel section."""

from collections.abc import Callable
from dataclasses import dataclass
from typing import Self

from matplotlib import pyplot as plt

from blueprints.materials.steel import SteelMaterial
from blueprints.structural_sections.cross_section_cornered import CircularCorneredCrossSection
from blueprints.structural_sections.cross_section_rectangle import RectangularCrossSection
from blueprints.structural_sections.steel.steel_cross_sections._steel_cross_section import CombinedSteelCrossSection
from blueprints.structural_sections.steel.steel_cross_sections.plotters.general_steel_plotter import plot_shapes
from blueprints.structural_sections.steel.steel_cross_sections.standard_profiles.rhs import RHS
from blueprints.structural_sections.steel.steel_cross_sections.standard_profiles.rhscf import RHSCF
from blueprints.structural_sections.steel.steel_cross_sections.standard_profiles.shs import SHS
from blueprints.structural_sections.steel.steel_cross_sections.standard_profiles.shscf import SHSCF
from blueprints.structural_sections.steel.steel_element import SteelElement
from blueprints.type_alias import MM


@dataclass(kw_only=True)
class RHSSteelProfile(CombinedSteelCrossSection):
    """Representation of an SHS or RHS steel section.

    Attributes
    ----------
    steel_material : SteelMaterial
        Steel material properties for the profile.
    total_width : MM
        The width of the profile [mm].
    total_height : MM
        The height of the profile [mm].
    left_wall_thickness : MM
        The thickness of the left wall [mm].
    right_wall_thickness : MM
        The thickness of the right wall [mm].
    top_wall_thickness : MM
        The thickness of the top wall [mm].
    bottom_wall_thickness : MM
        The thickness of the bottom wall [mm].
    top_right_inner_radius : MM | None
        The inner radius of the top right corner. Default is None, the corner radius is then taken as the thickness.
    top_left_inner_radius : MM | None
        The inner radius of the top left corner. Default is None, the corner radius is then taken as the thickness.
    bottom_right_inner_radius : MM | None
        The inner radius of the bottom right corner. Default is None, the corner radius is then taken as the thickness.
    bottom_left_inner_radius : MM | None
        The inner radius of the bottom left corner. Default is None, the corner radius is then taken as the thickness.
    top_right_outer_radius : MM | None
        The outer radius of the top right corner. Default is None, the corner radius is then taken as twice the thickness.
    top_left_outer_radius : MM | None
        The outer radius of the top left corner. Default is None, the corner radius is then taken as twice the thickness.
    bottom_right_outer_radius : MM | None
        The outer radius of the bottom right corner. Default is None, the corner radius is then taken as twice the thickness.
    bottom_left_outer_radius : MM | None
        The outer radius of the bottom left corner. Default is None, the corner radius is then taken as twice the thickness.
    name : str
        The name of the profile. Default is "RHS-Profile". If corrosion is applied, the name will include the corrosion value.
    """

    steel_material: SteelMaterial
    total_width: MM
    total_height: MM
    left_wall_thickness: MM
    right_wall_thickness: MM
    top_wall_thickness: MM
    bottom_wall_thickness: MM
    top_right_inner_radius: MM | None = None
    top_left_inner_radius: MM | None = None
    bottom_right_inner_radius: MM | None = None
    bottom_left_inner_radius: MM | None = None
    top_right_outer_radius: MM | None = None
    top_left_outer_radius: MM | None = None
    bottom_right_outer_radius: MM | None = None
    bottom_left_outer_radius: MM | None = None
    name: str = "RHS-Profile"

    def __post_init__(self) -> None:
        """Initialize the RHS- or SHS-profile steel section."""
        self.top_right_inner_radius = self.top_right_inner_radius if self.top_right_inner_radius is not None else self.top_wall_thickness
        self.top_left_inner_radius = self.top_left_inner_radius if self.top_left_inner_radius is not None else self.top_wall_thickness
        self.bottom_right_inner_radius = self.bottom_right_inner_radius if self.bottom_right_inner_radius is not None else self.bottom_wall_thickness
        self.bottom_left_inner_radius = self.bottom_left_inner_radius if self.bottom_left_inner_radius is not None else self.bottom_wall_thickness
        self.top_right_outer_radius = self.top_right_outer_radius if self.top_right_outer_radius is not None else 2 * self.top_wall_thickness
        self.top_left_outer_radius = self.top_left_outer_radius if self.top_left_outer_radius is not None else 2 * self.top_wall_thickness
        self.bottom_right_outer_radius = (
            self.bottom_right_outer_radius if self.bottom_right_outer_radius is not None else 2 * self.bottom_wall_thickness
        )
        self.bottom_left_outer_radius = self.bottom_left_outer_radius if self.bottom_left_outer_radius is not None else 2 * self.bottom_wall_thickness

        # calculate the lengths of the rectangular sections
        self.right_wall_height = (
            self.total_height - self.top_wall_thickness - self.bottom_wall_thickness - self.top_right_inner_radius - self.bottom_right_inner_radius
        )
        self.left_wall_height = (
            self.total_height - self.top_wall_thickness - self.bottom_wall_thickness - self.top_left_inner_radius - self.bottom_left_inner_radius
        )
        self.top_wall_width = (
            self.total_width - self.left_wall_thickness - self.right_wall_thickness - self.top_right_inner_radius - self.top_left_inner_radius
        )
        self.bottom_wall_width = (
            self.total_width - self.left_wall_thickness - self.right_wall_thickness - self.bottom_right_inner_radius - self.bottom_left_inner_radius
        )

        # Create the cross-sections for the flanges and web
        self.top_wall = RectangularCrossSection(
            name="Top Wall",
            width=self.top_wall_width,
            height=self.top_wall_thickness,
            x=(self.left_wall_thickness - self.right_wall_thickness + self.top_left_inner_radius - self.top_right_inner_radius) / 2,
            y=(self.total_height - self.top_wall_thickness) / 2,
        )
        self.bottom_wall = RectangularCrossSection(
            name="Bottom Wall",
            width=self.bottom_wall_width,
            height=self.bottom_wall_thickness,
            x=(self.left_wall_thickness - self.right_wall_thickness + self.bottom_left_inner_radius - self.bottom_right_inner_radius) / 2,
            y=-(self.total_height - self.bottom_wall_thickness) / 2,
        )
        self.left_wall = RectangularCrossSection(
            name="Left Wall",
            width=self.left_wall_thickness,
            height=self.left_wall_height,
            x=-(self.total_width - self.left_wall_thickness) / 2,
            y=-(self.top_wall_thickness - self.bottom_wall_thickness + self.top_left_inner_radius - self.bottom_left_inner_radius) / 2,
        )
        self.right_wall = RectangularCrossSection(
            name="Right Wall",
            width=self.right_wall_thickness,
            height=self.right_wall_height,
            x=(self.total_width - self.right_wall_thickness) / 2,
            y=-(self.top_wall_thickness - self.bottom_wall_thickness + self.top_right_inner_radius - self.bottom_right_inner_radius) / 2,
        )

        # Create the corner sections
        self.top_right_corner = CircularCorneredCrossSection(
            name="Top Right Corner",
            thickness_vertical=self.top_wall_thickness,
            thickness_horizontal=self.right_wall_thickness,
            inner_radius=self.top_right_inner_radius,
            outer_radius=self.top_right_outer_radius,
            x=self.total_width / 2 - self.right_wall_thickness - self.top_right_inner_radius,
            y=self.total_height / 2 - self.top_wall_thickness - self.top_right_inner_radius,
            corner_direction=0,
        )
        self.top_left_corner = CircularCorneredCrossSection(
            name="Top Left Corner",
            thickness_vertical=self.top_wall_thickness,
            thickness_horizontal=self.left_wall_thickness,
            inner_radius=self.top_left_inner_radius,
            outer_radius=self.top_left_outer_radius,
            x=-self.total_width / 2 + self.left_wall_thickness + self.top_left_inner_radius,
            y=self.total_height / 2 - self.top_wall_thickness - self.top_left_inner_radius,
            corner_direction=1,
        )
        self.bottom_right_corner = CircularCorneredCrossSection(
            name="Bottom Right Corner",
            thickness_vertical=self.bottom_wall_thickness,
            thickness_horizontal=self.right_wall_thickness,
            inner_radius=self.bottom_right_inner_radius,
            outer_radius=self.bottom_right_outer_radius,
            x=self.total_width / 2 - self.right_wall_thickness - self.bottom_right_inner_radius,
            y=-self.total_height / 2 + self.bottom_wall_thickness + self.bottom_right_inner_radius,
            corner_direction=3,
        )
        self.bottom_left_corner = CircularCorneredCrossSection(
            name="Bottom Left Corner",
            thickness_vertical=self.bottom_wall_thickness,
            thickness_horizontal=self.left_wall_thickness,
            inner_radius=self.bottom_left_inner_radius,
            outer_radius=self.bottom_left_outer_radius,
            x=-self.total_width / 2 + self.left_wall_thickness + self.bottom_left_inner_radius,
            y=-self.total_height / 2 + self.bottom_wall_thickness + self.bottom_left_inner_radius,
            corner_direction=2,
        )

        # Create the steel elements
        self.elements = [
            SteelElement(
                cross_section=self.top_wall,
                material=self.steel_material,
                nominal_thickness=self.top_wall_thickness,
            ),
            SteelElement(
                cross_section=self.bottom_wall,
                material=self.steel_material,
                nominal_thickness=self.bottom_wall_thickness,
            ),
            SteelElement(
                cross_section=self.left_wall,
                material=self.steel_material,
                nominal_thickness=self.left_wall_thickness,
            ),
            SteelElement(
                cross_section=self.right_wall,
                material=self.steel_material,
                nominal_thickness=self.right_wall_thickness,
            ),
            SteelElement(
                cross_section=self.top_right_corner,
                material=self.steel_material,
                nominal_thickness=self.top_wall_thickness,
            ),
            SteelElement(
                cross_section=self.top_left_corner,
                material=self.steel_material,
                nominal_thickness=self.top_wall_thickness,
            ),
            SteelElement(
                cross_section=self.bottom_right_corner,
                material=self.steel_material,
                nominal_thickness=self.bottom_wall_thickness,
            ),
            SteelElement(
                cross_section=self.bottom_left_corner,
                material=self.steel_material,
                nominal_thickness=self.bottom_wall_thickness,
            ),
        ]

    @classmethod
    def from_standard_profile(
        cls,
        profile: RHS | SHS | RHSCF | SHSCF,
        steel_material: SteelMaterial,
        corrosion_outside: MM = 0,
        corrosion_inside: MM = 0,
    ) -> Self:
        """Create an RHS- or SHS-profile from a set of standard profiles already defined in Blueprints.

        Blueprints offers standard profiles for RHS, SHS, RHSCF, and SHSCF. This method allows you to create an RHS/SHS-profile.

        Parameters
        ----------
        profile : RHS | SHS | RHSCF | SHSCF
            Any of the standard profiles defined in Blueprints.
        steel_material : SteelMaterial
            Steel material properties for the profile.
        corrosion_outside : MM, optional
            Corrosion thickness to be subtracted from the outer diameter [mm] (default: 0).
        corrosion_inside : MM, optional
            Corrosion thickness to be added to the inner diameter [mm] (default: 0).
        """
        total_width = profile.total_width - 2 * corrosion_outside
        total_height = profile.total_height - 2 * corrosion_outside

        top_wall_thickness = profile.thickness - corrosion_outside - corrosion_inside
        bottom_wall_thickness = profile.thickness - corrosion_outside - corrosion_inside
        left_wall_thickness = profile.thickness - corrosion_outside - corrosion_inside
        right_wall_thickness = profile.thickness - corrosion_outside - corrosion_inside

        top_right_inner_radius = profile.inner_radius + corrosion_inside
        top_left_inner_radius = profile.inner_radius + corrosion_inside
        bottom_right_inner_radius = profile.inner_radius + corrosion_inside
        bottom_left_inner_radius = profile.inner_radius + corrosion_inside
        top_right_outer_radius = profile.outer_radius - corrosion_outside
        top_left_outer_radius = profile.outer_radius - corrosion_outside
        bottom_right_outer_radius = profile.outer_radius - corrosion_outside
        bottom_left_outer_radius = profile.outer_radius - corrosion_outside

        if any(
            [
                top_wall_thickness < 1e-3,
                bottom_wall_thickness < 1e-3,
                left_wall_thickness < 1e-3,
                right_wall_thickness < 1e-3,
            ]
        ):
            raise ValueError("The profile has fully corroded.")

        name = profile.alias
        if corrosion_inside or corrosion_outside:
            name += (
                f" (corrosion {'' if not corrosion_inside else f'in: {corrosion_inside} mm'}"
                f"{', ' if corrosion_inside and corrosion_outside else ''}"
                f"{'' if not corrosion_outside else f'out: {corrosion_outside} mm'})"
            )

        return cls(
            steel_material=steel_material,
            total_width=total_width,
            total_height=total_height,
            left_wall_thickness=left_wall_thickness,
            right_wall_thickness=right_wall_thickness,
            top_wall_thickness=top_wall_thickness,
            bottom_wall_thickness=bottom_wall_thickness,
            top_right_inner_radius=top_right_inner_radius,
            top_left_inner_radius=top_left_inner_radius,
            bottom_right_inner_radius=bottom_right_inner_radius,
            bottom_left_inner_radius=bottom_left_inner_radius,
            top_right_outer_radius=top_right_outer_radius,
            top_left_outer_radius=top_left_outer_radius,
            bottom_right_outer_radius=bottom_right_outer_radius,
            bottom_left_outer_radius=bottom_left_outer_radius,
            name=name,
        )

    def plot(self, plotter: Callable[[CombinedSteelCrossSection], plt.Figure] | None = None, *args, **kwargs) -> plt.Figure:
        """Plot the cross-section. Making use of the standard plotter.

        Parameters
        ----------
        plotter : Callable[CombinedSteelCrossSection, plt.Figure] | None
            The plotter function to use. If None, the default Blueprints plotter for steel sections is used.
        *args
            Additional arguments passed to the plotter.
        **kwargs
            Additional keyword arguments passed to the plotter.
        """
        if plotter is None:
            plotter = plot_shapes
        return plotter(
            self,
            *args,
            **kwargs,
        )
